#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "rater/neural.h"
#include "auto-raters.h"

#define max(xVal,yVal) ( (xVal) > (yVal) ? (xVal) : (yVal) )

/*
 * Weights for list rater
 */
double *oneWeights;
double oneBias;
double *twoWeights;
double twoBias;
double *threeWeights;
double threeBias;
double *outputWeights;
double outputBiases[7];		  

/*
 * Defines the level spreads for the list auto rater
 *
 *   The spread of the top level integrator will be adjusted to accept
 *   the maximum number of inputs from level 3 given the maximimum sequence
 *   length is L_MAX_LENGTH.
 */
double L_ONE_SPREAD=8;
double L_ONE_DELTA=4;
double L_TWO_SPREAD=8;
double L_TWO_DELTA=4;
double L_THREE_SPREAD=4;
double L_THREE_DELTA=2;
double L_MAX_LENGTH=0;

/*
 * Weights for tournament rater
 */
double *aOneWeights;
double aOneBias;
double *aTwoWeights;
double aTwoBias;
double *aThreeWeights;
double aThreeBias;
double *aOutputWeights;

double *bOneWeights;
double bOneBias;
double *bTwoWeights;
double bTwoBias;
double *bThreeWeights;
double bThreeBias;
double *bOutputWeights;

double outputBias;		  

/*
 * Defines the level spreads for the tournament auto rater
 *
 *   The spread of the top level integrator will be adjusted to accept
 *   the maximum number of inputs from level 3 given the maximimum sequence
 *   length is T_MAX_LENGTH.
 */
double T_ONE_SPREAD=8;
double T_ONE_DELTA=4;
double T_TWO_SPREAD=8;
double T_TWO_DELTA=4;
double T_THREE_SPREAD=4;
double T_THREE_DELTA=2;
double T_MAX_LENGTH=0;

/*
 * InitListRater - initializes the list rater network using the parameters,
 *   weights and biases found in weightFileName.  Note that either the list
 *   rater or the tournament rater can be initialized, but not both.
 */
void InitListRater(const char *weightFileName) {

  FILE *weightFile;
  double outputLength;
  int i;

  /* open the file */
  weightFile=fopen(weightFileName, "r");
  
  /* load the spread and delta values */
  fread(&L_ONE_SPREAD, sizeof(double), 1, weightFile);
  fread(&L_ONE_DELTA, sizeof(double), 1, weightFile);
  fread(&L_TWO_SPREAD, sizeof(double), 1, weightFile);
  fread(&L_TWO_DELTA, sizeof(double), 1, weightFile);
  fread(&L_THREE_SPREAD, sizeof(double), 1, weightFile);
  fread(&L_THREE_DELTA, sizeof(double), 1, weightFile);
  fread(&L_MAX_LENGTH, sizeof(double), 1, weightFile);

  /* create the weight and bias arrays */
  oneWeights=(double *)malloc(L_ONE_SPREAD*sizeof(double));
  twoWeights=(double *)malloc(L_TWO_SPREAD*sizeof(double));
  threeWeights=(double *)malloc(L_THREE_SPREAD*sizeof(double));

  outputLength=max(1,ceil(1+(L_MAX_LENGTH-L_ONE_SPREAD)/L_ONE_DELTA));
  outputLength=max(1,ceil(1+(outputLength-L_TWO_SPREAD)/L_TWO_DELTA));
  outputLength=max(1,ceil(1+(outputLength-L_THREE_SPREAD)/L_THREE_DELTA));
  
  outputLength*=7;
  
  outputWeights=(double *)malloc(outputLength*sizeof(double));

  /* load the weights and biases */
  fread(oneWeights, sizeof(double), L_ONE_SPREAD, weightFile);
  fread(&oneBias, sizeof(double), 1, weightFile);

  fread(twoWeights, sizeof(double), L_TWO_SPREAD, weightFile);
  fread(&twoBias, sizeof(double), 1, weightFile);

  fread(threeWeights, sizeof(double), L_THREE_SPREAD, weightFile);
  fread(&threeBias, sizeof(double), 1, weightFile);

  fread(outputWeights, sizeof(double), (size_t)outputLength, weightFile);
  fread(outputBiases, sizeof(double), 7, weightFile);

  /* close the file */
  fclose(weightFile);

} /* end of InitListRater */

/*
 * InitTournRater - initializes the list rater network using the parameters,
 *   weights and biases found in weightFileName.  Note that either the list
 *   rater or the tournament rater can be initialized, but not both.
 */
void InitTournRater(const char *weightFileName) {

  FILE *weightFile;
  double outputLength;
  int i;

  /* open the file */
  weightFile=fopen(weightFileName, "r");
  
  /* load the spread and delta values */
  fread(&T_ONE_SPREAD, sizeof(double), 1, weightFile);
  fread(&T_ONE_DELTA, sizeof(double), 1, weightFile);
  fread(&T_TWO_SPREAD, sizeof(double), 1, weightFile);
  fread(&T_TWO_DELTA, sizeof(double), 1, weightFile);
  fread(&T_THREE_SPREAD, sizeof(double), 1, weightFile);
  fread(&T_THREE_DELTA, sizeof(double), 1, weightFile);
  fread(&T_MAX_LENGTH, sizeof(double), 1, weightFile);

  /* create the weight and bias arrays */
  aOneWeights=(double *)malloc(T_ONE_SPREAD*sizeof(double));
  aTwoWeights=(double *)malloc(T_TWO_SPREAD*sizeof(double));
  aThreeWeights=(double *)malloc(T_THREE_SPREAD*sizeof(double));

  bOneWeights=(double *)malloc(T_ONE_SPREAD*sizeof(double));
  bTwoWeights=(double *)malloc(T_TWO_SPREAD*sizeof(double));
  bThreeWeights=(double *)malloc(T_THREE_SPREAD*sizeof(double));

  outputLength=max(1,ceil(1+(T_MAX_LENGTH-T_ONE_SPREAD)/T_ONE_DELTA));
  outputLength=max(1,ceil(1+(outputLength-T_TWO_SPREAD)/T_TWO_DELTA));
  outputLength=max(1,ceil(1+(outputLength-T_THREE_SPREAD)/T_THREE_DELTA));

  aOutputWeights=(double *)malloc(outputLength*sizeof(double));
  bOutputWeights=(double *)malloc(outputLength*sizeof(double));

  /* load the weights and biases */
  fread(aOneWeights, sizeof(double), T_ONE_SPREAD, weightFile);
  fread(&aOneBias, sizeof(double), 1, weightFile);

  fread(bOneWeights, sizeof(double), T_ONE_SPREAD, weightFile);
  fread(&bOneBias, sizeof(double), 1, weightFile);

  fread(aTwoWeights, sizeof(double), T_TWO_SPREAD, weightFile);
  fread(&aTwoBias, sizeof(double), 1, weightFile);

  fread(bTwoWeights, sizeof(double), T_TWO_SPREAD, weightFile);
  fread(&bTwoBias, sizeof(double), 1, weightFile);

  fread(aThreeWeights, sizeof(double), T_THREE_SPREAD, weightFile);
  fread(&aThreeBias, sizeof(double), 1, weightFile);

  fread(bThreeWeights, sizeof(double), T_THREE_SPREAD, weightFile);
  fread(&bThreeBias, sizeof(double), 1, weightFile);

  fread(aOutputWeights, sizeof(double), (size_t)outputLength, weightFile);
  fread(bOutputWeights, sizeof(double), (size_t)outputLength, weightFile);
  fread(&outputBias, sizeof(double), 1, weightFile);

  /* close the file */
  fclose(weightFile);
  
} /* end of InitTournRater */

/*
 * tAutoRate - gives the neural nets answer to whether sequence1 is worse than
 *   sequence2.  A result of less than 0.5 indicates that it is worse, and
 *   greater than 0.5 indicates that sequence1 is better.
 */
double tAutoRate(const char *sequence1, const char *sequence2) {

  Neuron **aInputLayer;
  Neuron **bInputLayer;
  Neuron **aFirstLayer;
  Neuron **bFirstLayer;
  Neuron **aSecondLayer;
  Neuron **bSecondLayer;
  Neuron **aThirdLayer;
  Neuron **bThirdLayer;
  Neuron *output;

  int prevLayerSize, layerSize;
  double theResult;
  int i,j;
  
  /*
   * Create the 'A' network layers (for individual 1)
   */
  
  /* input layer */
  layerSize=strlen(sequence1);
  aInputLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    aInputLayer[i]=createNeuron(0);

  /* first layer */
  layerSize=max(1,ceil(1+(layerSize-T_ONE_SPREAD)/T_ONE_DELTA));
  aFirstLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    aFirstLayer[i]=createNeuron(&aOneBias);

  /* second layer */
  layerSize=max(1,ceil(1+(layerSize-T_TWO_SPREAD)/T_TWO_DELTA));
  aSecondLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    aSecondLayer[i]=createNeuron(&aTwoBias);

  /* third layer */
  layerSize=max(1,ceil(1+(layerSize-T_THREE_SPREAD)/T_THREE_DELTA));
  aThirdLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    aThirdLayer[i]=createNeuron(&aThreeBias);

  /*
   * Create the 'B' network layers (for individual 2)
   */
  
  /* input layer */
  layerSize=strlen(sequence2);
  bInputLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    bInputLayer[i]=createNeuron(0);

  /* first layer */
  layerSize=max(1,ceil(1+(layerSize-T_ONE_SPREAD)/T_ONE_DELTA));
  bFirstLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    bFirstLayer[i]=createNeuron(&bOneBias);

  /* second layer */
  layerSize=max(1,ceil(1+(layerSize-T_TWO_SPREAD)/T_TWO_DELTA));
  bSecondLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    bSecondLayer[i]=createNeuron(&bTwoBias);

  /* third layer */
  layerSize=max(1,ceil(1+(layerSize-T_THREE_SPREAD)/T_THREE_DELTA));
  bThirdLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    bThirdLayer[i]=createNeuron(&bThreeBias);

  /*
   * Create the output node
   */
  output=createNeuron(&outputBias);

  /*
   * Now make connections between the layers of side 'A'
   */

  /* connections from inputs to first layer nodes */
  prevLayerSize=layerSize=strlen(sequence1);
  layerSize=max(1,ceil(1+(layerSize-T_ONE_SPREAD)/T_ONE_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<T_ONE_SPREAD;j++)
      if(T_ONE_DELTA*i+j < prevLayerSize)
	connectNeurons(aInputLayer[(int)T_ONE_DELTA*i+j], aFirstLayer[i],
		       &(aOneWeights[j]));

  /* connections from first layer to second layer nodes */
  prevLayerSize=layerSize;
  layerSize=max(1,ceil(1+(prevLayerSize-T_TWO_SPREAD)/T_TWO_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<T_TWO_SPREAD;j++)
      if(T_TWO_DELTA*i+j < prevLayerSize)
	connectNeurons(aFirstLayer[(int)T_TWO_DELTA*i+j], aSecondLayer[i],
		       &(aTwoWeights[j]));

  /* connections from second layer to third layer nodes */
  prevLayerSize=layerSize;
  layerSize=max(1,ceil(1+(prevLayerSize-T_THREE_SPREAD)/T_THREE_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<T_THREE_SPREAD;j++)
      if(T_THREE_DELTA*i+j < prevLayerSize)
	connectNeurons(aSecondLayer[(int)T_THREE_DELTA*i+j], aThirdLayer[i],
		       &(aThreeWeights[j]));

  /* connections from third layer nodes to output node */
  for(i=0;i<layerSize;i++)
      connectNeurons(aThirdLayer[i], output,
		     &(aOutputWeights[i]));

  /*
   * Now make connections between the layers of side 'B'
   */

  /* connections from inputs to first layer nodes */
  prevLayerSize=layerSize=strlen(sequence2);
  layerSize=max(1,ceil(1+(layerSize-T_ONE_SPREAD)/T_ONE_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<T_ONE_SPREAD;j++)
      if(T_ONE_DELTA*i+j < prevLayerSize)
	connectNeurons(bInputLayer[(int)T_ONE_DELTA*i+j], bFirstLayer[i],
		       &(bOneWeights[j]));

  /* connections from first layer to second layer nodes */
  prevLayerSize=layerSize;
  layerSize=max(1,ceil(1+(prevLayerSize-T_TWO_SPREAD)/T_TWO_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<T_TWO_SPREAD;j++)
      if(T_TWO_DELTA*i+j < prevLayerSize)
	connectNeurons(bFirstLayer[(int)T_TWO_DELTA*i+j], bSecondLayer[i],
		       &(bTwoWeights[j]));

  /* connections from second layer to third layer nodes */
  prevLayerSize=layerSize;
  layerSize=max(1,ceil(1+(prevLayerSize-T_THREE_SPREAD)/T_THREE_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<T_THREE_SPREAD;j++)
      if(T_THREE_DELTA*i+j < prevLayerSize)
	connectNeurons(bSecondLayer[(int)T_THREE_DELTA*i+j], bThirdLayer[i],
		       &(bThreeWeights[j]));

  /* connections from third layer nodes to output node */
  for(i=0;i<layerSize;i++)
      connectNeurons(bThirdLayer[i], output,
		     &(bOutputWeights[i]));

  /*
   * Now load the noteSequences values into the inputs
   */
  layerSize=strlen(sequence1);
  for(i=0;i<layerSize;i++)
    setActivation(aInputLayer[i],
		  (sequence1[i] > 72 ?
		   -1 :
		   sequence1[i]/72));

  layerSize=strlen(sequence2);
  for(i=0;i<layerSize;i++)
    setActivation(bInputLayer[i],
		  (sequence2[i] > 72 ?
		   -1 :
		   sequence2[i]/72));
    
  /*
   * Now get the answer
   */
  theResult=findOutput(output);
  
  /*
   * Now destroy the network, layer by layer
   */

  /*
   * Destroy side 'A'
   */
  
  /* input layer */
  layerSize=strlen(sequence1);
  for(i=0;i<layerSize;i++)
    freeNeuron(aInputLayer[i]);
  free(aInputLayer);

  /* first layer */
  layerSize=max(1,ceil(1+(layerSize-T_ONE_SPREAD)/T_ONE_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(aFirstLayer[i]);
  free(aFirstLayer);

  /* second layer */
  layerSize=max(1,ceil(1+(layerSize-T_TWO_SPREAD)/T_TWO_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(aSecondLayer[i]);
  free(aSecondLayer);
  
  /* third layer */
  layerSize=max(1,ceil(1+(layerSize-T_THREE_SPREAD)/T_THREE_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(aThirdLayer[i]);
  free(aThirdLayer);

  /*
   * Destroy side 'B'
   */
  
  /* input layer */
  layerSize=strlen(sequence2);
  for(i=0;i<layerSize;i++)
    freeNeuron(bInputLayer[i]);
  free(bInputLayer);

  /* first layer */
  layerSize=max(1,ceil(1+(layerSize-T_ONE_SPREAD)/T_ONE_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(bFirstLayer[i]);
  free(bFirstLayer);

  /* second layer */
  layerSize=max(1,ceil(1+(layerSize-T_TWO_SPREAD)/T_TWO_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(bSecondLayer[i]);
  free(bSecondLayer);
  
  /* third layer */
  layerSize=max(1,ceil(1+(layerSize-T_THREE_SPREAD)/T_THREE_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(bThirdLayer[i]);
  free(bThirdLayer);

  /*
   * free the output node
   */
  freeNeuron(output);

  /*
   * now return with the result
   */
  return theResult;
  
} /* end of tAutoRate */

/*
 * lAutoRate - gives the neural nets rating for aSequence.  The answer varies
 *   varies between 0 and 1, with 1 being the best.
 */
double lAutoRate(const char *aSequence) {

  Neuron **inputLayer;
  Neuron **firstLayer;
  Neuron **secondLayer;
  Neuron **thirdLayer;
  Neuron *output;

  int prevLayerSize, layerSize;
  double theResult;
  int i,j;
  
  /*
   * Create the network layers
   */

  /* input layer */
  layerSize=strlen(aSequence);
  inputLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    inputLayer[i]=createNeuron(0);

  /* first layer */
  layerSize=max(1,ceil(1+(layerSize-L_ONE_SPREAD)/L_ONE_DELTA));
  firstLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    firstLayer[i]=createNeuron(&oneBias);

  /* second layer */
  layerSize=max(1,ceil(1+(layerSize-L_TWO_SPREAD)/L_TWO_DELTA));
  secondLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    secondLayer[i]=createNeuron(&twoBias);

  /* third layer */
  layerSize=max(1,ceil(1+(layerSize-L_THREE_SPREAD)/L_THREE_DELTA));
  thirdLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    thirdLayer[i]=createNeuron(&threeBias);

  /* output nodes */
  output=createNeuron(&(outputBiases[0]));

  /*
   * Now make connections between the layers
   */

  /* connections from inputs to first layer nodes */
  prevLayerSize=layerSize=strlen(aSequence);
  layerSize=max(1,ceil(1+(layerSize-L_ONE_SPREAD)/L_ONE_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<L_ONE_SPREAD;j++)
      if(L_ONE_DELTA*i+j < prevLayerSize)
	connectNeurons(inputLayer[(int)L_ONE_DELTA*i+j], firstLayer[i],
		       &(oneWeights[j]));

  /* connections from first layer to second layer nodes */
  prevLayerSize=layerSize;
  layerSize=max(1,ceil(1+(prevLayerSize-L_TWO_SPREAD)/L_TWO_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<L_TWO_SPREAD;j++)
      if(L_TWO_DELTA*i+j < prevLayerSize)
	connectNeurons(firstLayer[(int)L_TWO_DELTA*i+j], secondLayer[i],
		       &(twoWeights[j]));

  /* connections from second layer to third layer nodes */
  prevLayerSize=layerSize;
  layerSize=max(1,ceil(1+(prevLayerSize-L_THREE_SPREAD)/L_THREE_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<L_THREE_SPREAD;j++)
      if(L_THREE_DELTA*i+j < prevLayerSize)
	connectNeurons(secondLayer[(int)L_THREE_DELTA*i+j], thirdLayer[i],
		       &(threeWeights[j]));

  /* connections from third layer nodes to output node */
  for(i=0;i<layerSize;i++)
      connectNeurons(thirdLayer[i], output,
		     &(outputWeights[i]));

  /*
   * Now load the noteSequence values into the inputs
   */
  layerSize=strlen(aSequence);
  for(i=0;i<layerSize;i++)
    setActivation(inputLayer[i],
		  (aSequence[i] > 72 ?
		   -1 :
		   aSequence[i]/72));
    
  /*
   * Now get the rating
   */
  theResult=findOutput(output);
  
  /*
   * Now destroy the network, layer by layer
   */

  /* input layer */
  layerSize=strlen(aSequence);
  for(i=0;i<layerSize;i++)
    freeNeuron(inputLayer[i]);
  free(inputLayer);

  /* first layer */
  layerSize=max(1,ceil(1+(layerSize-L_ONE_SPREAD)/L_ONE_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(firstLayer[i]);
  free(firstLayer);

  /* second layer */
  layerSize=max(1,ceil(1+(layerSize-L_TWO_SPREAD)/L_TWO_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(secondLayer[i]);
  free(secondLayer);
  
  /* third layer */
  layerSize=max(1,ceil(1+(layerSize-L_THREE_SPREAD)/L_THREE_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(thirdLayer[i]);
  free(thirdLayer);

  /* output node */
  freeNeuron(output);

  /*
   * now return with the result
   */
  return theResult;
  
} /* end of lAutoRate */

/*
 * uninitRaters - frees all memory used by the raters
 */
void uninitRaters() {

  /* free the list rater stuff */
  free(oneWeights);
  free(twoWeights);
  free(threeWeights);
  free(outputWeights);
  
  /* free the tournament rater stuff */
  free(aOneWeights);
  free(aTwoWeights);
  free(aThreeWeights);
  free(bOneWeights);
  free(bTwoWeights);
  free(bThreeWeights);
  free(aOutputWeights);
  free(bOutputWeights);

} /* end of uninitRaters */
