#include <string.h>
#include "cache.h"
#include "lilgp.h"

/*
 * cache.h cache.c - these files cache old noteSequences and their ratings.
 *   The cache can be polled to see if a noteSequence has been rated before,
 *   and if so, what it was rated.
 */
typedef struct
{
  char *noteSequence;
  double rating;
  
} CacheEntry;

CacheEntry *theCache;
int initialized=0;
int currentSize=0;
int maxSize=0;

/*
 * initCache - initializes the cache for a run with numGen generations and
 *   numInd individuals per generation
 */
void
initCache(int numGen, int numInd) {

  if(initialized) {
    resetCache();
    FREE(theCache);
  }
  
  maxSize=(numGen+1)*numInd;
  theCache=(CacheEntry *)MALLOC(maxSize*sizeof(CacheEntry));

  initialized=1;
  currentSize=0;
  
} /* end of initCache */

/*
 * resetCache - clears the current cached entries
 */
void
resetCache() {

  int i;

  if(!initialized) return;
  
  for(i=0;i<currentSize;i++)
    free(theCache[i].noteSequence);

  currentSize=0;
    
} /* end of resetCache */

/*
 * closeCache - deallocates all memory used by the cache.  The cache must
 *   then be reinitialized before it can be used again.
 */
void
closeCache() {

  if(!initialized) return;

  resetCache();

  FREE(theCache);

  initialized=maxSize=currentSize=0;

} /* end of closeCache */

/*
 * queryCache - checks to see if the noteSequence has been rated before.  If
 *   so, the returned value is the previous rating.  Otherwise the value
 *   returned is -1.
 */
double
queryCache(const char *noteSequence) {

  int i;

  if(!initialized) return -1;
  
  for(i=0;i<currentSize;i++)
    if(strcmp(noteSequence, theCache[i].noteSequence)==0)
      return theCache[i].rating;

  return -1;
  
} /* end of queryCache */

/*
 * updateCache - notifies the cache that the given noteSequence has been
 *   assigned the given new rating.  The noteSequence need not have been
 *   in the cache before.
 */
void
updateCache(const char *noteSequence, double rating) {

  int i;

  if(!initialized) return;

  for(i=0;i<currentSize;i++)
    if(strcmp(noteSequence, theCache[i].noteSequence)==0) {
      theCache[i].rating=rating;
      return;
    }

  /* not in cache, create new entry */
  if(i==currentSize) {

    if(currentSize==maxSize) return;

    theCache[currentSize].noteSequence=strdup(noteSequence);
    theCache[currentSize++].rating=rating;
    
  }
  
} /* end of updateCache */






