#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include "neural.h"

#define square(xVal) (xVal)*(xVal)

Neuron *inputs[4];
Neuron *hidden[2];
Neuron *outputs[4];

/*
 * ReadSysClock - returns the time in seconds and fractions of seconds
 *     as a double (for use only for calculating deltas)
 */
double
ReadSysClock ()
{

    struct timeval tv;

    (void) gettimeofday(&tv, NULL);
    return (tv.tv_sec +tv.tv_usec / 1000000.0);

} /* end of ReadSysClock */

/*
 * trialTrain - back propagates and calculates deltas for the given input
 *   and output, and returns the sum of the square errors across output
 *   nodes.  If applyErr is TRUE, it changes the weights.
 */
double trialTrain(double inputVals[4], double outputVals[4], int applyErr) {

  int i;
  double sumSqrErr=0;

  /* reset the network */
  for(i=0;i<4;i++)
    resetNet(outputs[i]);

  /* load the input values */
  for(i=0;i<4;i++)
    setActivation(inputs[i], inputVals[i]);

  /* back prop and get error for each output */
  for(i=0;i<4;i++)
    sumSqrErr+= square(backProp(outputs[i], outputVals[i]));

  /* calculate change in weights for this input set */
  for(i=0;i<4;i++)
    calcDeltas(outputs[i]);      

  /* if requested, change the weights */
  if(applyErr)
    for(i=0;i<4;i++)
      applyDeltas(outputs[i]);

  return sumSqrErr;

} /* end of trialTrain */

/*
 * trialTest - prints the output values yielded from the net with the given
 *   input values.
 */
void trialTest(double inputVals[4]) {

  int i;

  /* reset the network */
  for(i=0;i<4;i++)
    resetNet(outputs[i]);

  /* load the input values */
  for(i=0;i<4;i++)
    setActivation(inputs[i], inputVals[i]);

  /* now print the results */
  printf("Input pattern %g%g%g%g yields:\n\t%8.3g\t%8.3g\t%8.3g\t%8.3g\n",
	 inputVals[0], inputVals[1], inputVals[2], inputVals[3],
	 findOutput(outputs[0]), findOutput(outputs[1]),
	 findOutput(outputs[2]), findOutput(outputs[3]));

} /* end of trialTest */

int main(int argc, char *argv[]) {

  int i,j;
  int iterations=0;
  int alwaysApplyErrDeltas=0;
  double sumSumSqrErr;
  double maxErr=0.1;
  double startTime;
  double lrate=0.5;
  double momentum=0;
  
  /* some small random weights and biases to start with */
  double hiddenBiases[2]= { -1.2, -0.9 };
  double outputBiases[4]= { -2.3, 1.2, -1.4, 0.5 };
  double hiddenWeights[8]= { 0.01, 0.02, 0.03, 0.04, 0.01, 0.02, 0.03, 0.04 };
  double outputWeights[8]= { 0.01, 0.02, 0.03, 0.04, 0.01, 0.02, 0.03, 0.04 };

  /* array for input and output values */
  double values[4];
  
  /* init with command line params */
  for(i=1;i<argc;i++) {
    
    if(!strcmp("-maxErr",argv[i])) {
      maxErr=atof(argv[++i]);
    } else if(!strcmp("-A",argv[i])) {
      alwaysApplyErrDeltas=1;
    } else if(!strcmp("-lrate",argv[i])) {
      lrate=atof(argv[++i]);
    } else if(!strcmp("-momentum",argv[i])) {
      momentum=atof(argv[++i]);
    } else {
      printf("\nSyntax: %s [-A] [-maxErr <val>] [-lrate <val>] "
	     "[-momentum val]\n\n"
	     "\t-maxErr\t\tspecifies the target error value\n"
	     "\t-lrate\t\tspecifies the learning rate\n"
	     "\t-momentum\tspecifies the momentum for learning\n"
	     "\t-A\t\tchange weights after each trial\n\n", argv[0]);
      exit(1);
    }

  }

  printf("\nRunning Correlate Problem with:\n\tMaxErr=%g\tlrate=%g\t\t"
	 "momentum=%g\n\tError Deltas applied once per %s\n\n", maxErr, lrate,
	 momentum, (alwaysApplyErrDeltas?"trial.":"trial set."));
  setLearnParams(lrate, momentum, lrate, momentum);
  
  /* create the neurons */
  for(i=0;i<4;i++) {
    inputs[i]=createNeuron(0);
    outputs[i]=createNeuron(&(outputBiases[i]));
    if(i<2)
      hidden[i]=createNeuron(&(hiddenBiases[i]));
  }
    
  /* create the connections */
  
  /* inputs to hidden nodes */
  for(i=0;i<4;i++)
    for(j=0;j<2;j++)
      connectNeurons(inputs[i], hidden[j], &(hiddenWeights[2*i+j]));

  /* hidden nodes to outputs */
  for(i=0;i<4;i++)
    for(j=0;j<2;j++)
      connectNeurons(hidden[j], outputs[i], &(outputWeights[2*i+j]));
    
  /* Now train the network until error falls below maxErr */
  startTime=ReadSysClock();
  do {

    /* reset sumErr */
    sumSumSqrErr=0;
    
    /* first do '1000 -> 1000' */
    values[0]=1; values[1]=0; values[2]=0; values[3]=0;
    sumSumSqrErr+=trialTrain(values, values, alwaysApplyErrDeltas);

    /* now do '0100 -> 0100' */
    values[0]=0; values[1]=1; values[2]=0; values[3]=0;
    sumSumSqrErr+=trialTrain(values, values, alwaysApplyErrDeltas);

    /* now do '0010 -> 0010' */
    values[0]=0; values[1]=0; values[2]=1; values[3]=0;
    sumSumSqrErr+=trialTrain(values, values, alwaysApplyErrDeltas);

    /* now do '0001 -> 0001' */
    values[0]=0; values[1]=0; values[2]=0; values[3]=1;
    sumSumSqrErr+=trialTrain(values, values, 1);

    iterations++;
    
  } while(sumSumSqrErr>maxErr);

  /* print run statistics */
  printf("Run took %g seconds\n\n", ReadSysClock() - startTime);
  printf("Finished in %d iterations with a sum squared error of %g\n\n",
	 iterations, sumSumSqrErr);
  
  /* now print the network outputs for the various patterns */
  
  /* first print '1000 -> 1000' */
  values[0]=1; values[1]=0; values[2]=0; values[3]=0;
  trialTest(values);
   
  /* now print '0100 -> 0100' */
  values[0]=0; values[1]=1; values[2]=0; values[3]=0;
  trialTest(values);

  /* now print '0010 -> 0010' */
  values[0]=0; values[1]=0; values[2]=1; values[3]=0;
  trialTest(values);

  /* now print '0001 -> 0001' */
  values[0]=0; values[1]=0; values[2]=0; values[3]=1;
  trialTest(values);

  printf("\n");
  
  return 0;
  
}




