#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include "neural.h"

#define square(xVal) (xVal)*(xVal)
#define max(xVal,yVal) ( (xVal) > (yVal) ? (xVal) : (yVal) )
#define min(xVal,yVal) ( (xVal) < (yVal) ? (xVal) : (yVal) )

/* stop criterion types */
#define AVG 0
#define MAX 1
#define ITER 2

/* network configurations */
#define ONE_OUTPUT 0
#define SEVEN_OUTPUT 1

typedef struct {

  char *noteSequence;
  double rating;     /* note that this is r_fitness/100 */
  
} RatingPair;

double *oneWeights;
double oneBias;
double *twoWeights;
double twoBias;
double *threeWeights;
double threeBias;
double *outputWeights;
double outputBiases[7];		  

/*
 * Defines the level spreads for the list auto rater
 *
 *   The spread of the top level integrator will be adjusted to accept
 *   the maximum number of inputs from level 3 given the maximimum sequence
 *   length is L_MAX_LENGTH.
 */
double L_ONE_SPREAD=8;
double L_ONE_DELTA=4;
double L_TWO_SPREAD=8;
double L_TWO_DELTA=4;
double L_THREE_SPREAD=4;
double L_THREE_DELTA=2;
double L_MAX_LENGTH=0;

/*
 * ReadSysClock - returns the time in seconds and fractions of seconds
 *     as a double (for use only for calculating deltas)
 */
double
ReadSysClock ()
{

    struct timeval tv;

    (void) gettimeofday(&tv, NULL);
    return (tv.tv_sec +tv.tv_usec / 1000000.0);

} /* end of ReadSysClock */

/*
 * initNetVals - allocates the weight arrays, and initializes weighs and
 *   biases to small random values
 */
void initNetVals() {

  double outputLength;
  int i;
  
  /* create the weight and bias arrays */
  oneWeights=(double *)malloc(L_ONE_SPREAD*sizeof(double));
  twoWeights=(double *)malloc(L_TWO_SPREAD*sizeof(double));
  threeWeights=(double *)malloc(L_THREE_SPREAD*sizeof(double));

  outputLength=max(1,ceil(1+(L_MAX_LENGTH-L_ONE_SPREAD)/L_ONE_DELTA));
  outputLength=max(1,ceil(1+(outputLength-L_TWO_SPREAD)/L_TWO_DELTA));
  outputLength=max(1,ceil(1+(outputLength-L_THREE_SPREAD)/L_THREE_DELTA));

  outputLength*=7;
  
  outputWeights=(double *)malloc(outputLength*sizeof(double));

  /* fill the weight arrays and biases with random values */
  for(i=0;i<L_ONE_SPREAD;i++)
    oneWeights[i]=( drand48() > 0.5 ? 1 : -1)*drand48();
  for(i=0;i<L_TWO_SPREAD;i++)
    twoWeights[i]=( drand48() > 0.5 ? 1 : -1)*drand48();
  for(i=0;i<L_THREE_SPREAD;i++)
    threeWeights[i]=( drand48() > 0.5 ? 1 : -1)*drand48();
  for(i=0;i<outputLength;i++)
    outputWeights[i]=( drand48() > 0.5 ? 1 : -1)*drand48();

  oneBias=( drand48() > 0.5 ? 1 : -1)*drand48();
  twoBias=( drand48() > 0.5 ? 1 : -1)*drand48();
  threeBias=( drand48() > 0.5 ? 1 : -1)*drand48();

  for(i=0;i<7;i++)
    outputBiases[i]=( drand48() > 0.5 ? 1 : -1)*drand48();

} /* end of initNetVals */

/*
 * fwriteNetState - writes all of the networks weights and biases into
 *   the given file
 */
void fwriteNetState(FILE *outputFile) {

  double outputLength;
  
  fwrite(oneWeights, sizeof(double), L_ONE_SPREAD, outputFile);
  fwrite(&oneBias, sizeof(double), 1, outputFile);

  fwrite(twoWeights, sizeof(double), L_TWO_SPREAD, outputFile);
  fwrite(&twoBias, sizeof(double), 1, outputFile);

  fwrite(threeWeights, sizeof(double), L_THREE_SPREAD, outputFile);
  fwrite(&threeBias, sizeof(double), 1, outputFile);

  outputLength=max(1,ceil(1+(L_MAX_LENGTH-L_ONE_SPREAD)/L_ONE_DELTA));
  outputLength=max(1,ceil(1+(outputLength-L_TWO_SPREAD)/L_TWO_DELTA));
  outputLength=max(1,ceil(1+(outputLength-L_THREE_SPREAD)/L_THREE_DELTA));

  outputLength*=7;
    
  fwrite(outputWeights, sizeof(double), (size_t)outputLength, outputFile);
  fwrite(outputBiases, sizeof(double), 7, outputFile);

} /* end of fwriteNetState */

/*
 * loadDataSet - loads the noteSequences and ratings in the file named
 *   'setFileName' into the array 'theSet'.  The set is allocated to fit
 *   the number of individuals in the file.  The last element of the set
 *   array has rating set to '-1' and indicates the end of the array.
 *   The function returns the number of individuals in the set.
 */
int loadDataSet(char *setFileName, RatingPair **theSet) {

  int i;
  int numInds;
  int temp;
  FILE *setFile;
  RatingPair *newArray;
  
  /* open the file */
  setFile=fopen(setFileName, "r");

  /* find out how many individuals in the set */
  fread(&numInds, sizeof(int), 1, setFile);

  /* allocate the array */
  newArray=(RatingPair *)malloc((numInds+1)*sizeof(RatingPair));
  
  /* now loop through loading individuals */
  for(i=0;i<numInds;i++) {

      /* read size of sequence */
      fread(&temp, sizeof(int), 1, setFile);
      
      /* allocate and read note sequence */
      newArray[i].noteSequence=(char *)malloc(temp*sizeof(char));
      fread(newArray[i].noteSequence, sizeof(char), temp, setFile);

      /* read rating */
      fread(&(newArray[i].rating), sizeof(double), 1, setFile);

  }

  /* close the file */
  fclose(setFile);
  
  /* set the end of array */
  newArray[i].rating=-1;

  /* set the array pointer for return */
  *theSet=newArray;
  
  return numInds;

} /* end of loadDataSet */

/*
 * trialNet7 - creates a network for the noteSequence in 'example', and then
 *   applies it to the 'notesequence', returning the squared error between
 *   the nets output and the rating in 'example'.  If 'train' is 1, it also
 *   trains the net for the noteSequence and rating in 'example', thus
 *   adjusting the weights and biases. 'decError' is the returned RMS
 *   difference between the 7-bit scaled output and the rating.
 *
 *   trialNet7 constructs a network with 7 output nodes which are treated
 *   as a 7-bit binary number.  The rating (1-100) is scaled to this range
 *   for comparisons.
 */
double trialNet7(RatingPair *example, int train, double *decError) {

  Neuron **inputLayer;
  Neuron **firstLayer;
  Neuron **secondLayer;
  Neuron **thirdLayer;
  Neuron *outputLayer[7];

  double sumSqrErr=0;
  int scaleRating;
  int decimalValue=0;
  int decimalError;
  int prevLayerSize, layerSize;
  int i,j;
  
  /*
   * Create the network layers
   */

  /* input layer */
  layerSize=strlen(example->noteSequence);
  inputLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    inputLayer[i]=createNeuron(0);

  /* first layer */
  layerSize=max(1,ceil(1+(layerSize-L_ONE_SPREAD)/L_ONE_DELTA));
  firstLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    firstLayer[i]=createNeuron(&oneBias);

  /* second layer */
  layerSize=max(1,ceil(1+(layerSize-L_TWO_SPREAD)/L_TWO_DELTA));
  secondLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    secondLayer[i]=createNeuron(&twoBias);

  /* third layer */
  layerSize=max(1,ceil(1+(layerSize-L_THREE_SPREAD)/L_THREE_DELTA));
  thirdLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    thirdLayer[i]=createNeuron(&threeBias);

  /* output nodes */
  for(i=0;i<7;i++)
    outputLayer[i]=createNeuron(&(outputBiases[i]));

  /*
   * Now make connections between the layers
   */

  /* connections from inputs to first layer nodes */
  prevLayerSize=layerSize=strlen(example->noteSequence);
  layerSize=max(1,ceil(1+(layerSize-L_ONE_SPREAD)/L_ONE_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<L_ONE_SPREAD;j++)
      if(L_ONE_DELTA*i+j < prevLayerSize)
	connectNeurons(inputLayer[(int)L_ONE_DELTA*i+j], firstLayer[i],
		       &(oneWeights[j]));

  /* connections from first layer to second layer nodes */
  prevLayerSize=layerSize;
  layerSize=max(1,ceil(1+(prevLayerSize-L_TWO_SPREAD)/L_TWO_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<L_TWO_SPREAD;j++)
      if(L_TWO_DELTA*i+j < prevLayerSize)
	connectNeurons(firstLayer[(int)L_TWO_DELTA*i+j], secondLayer[i],
		       &(twoWeights[j]));

  /* connections from second layer to third layer nodes */
  prevLayerSize=layerSize;
  layerSize=max(1,ceil(1+(prevLayerSize-L_THREE_SPREAD)/L_THREE_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<L_THREE_SPREAD;j++)
      if(L_THREE_DELTA*i+j < prevLayerSize)
	connectNeurons(secondLayer[(int)L_THREE_DELTA*i+j], thirdLayer[i],
		       &(threeWeights[j]));

  /* connections from third layer nodes to output node */
  for(j=0;j<7;j++)
    for(i=0;i<layerSize;i++)
      connectNeurons(thirdLayer[i], outputLayer[j],
		     &(outputWeights[layerSize*j+i]));

  /*
   * Now load the noteSequence values into the inputs
   */
  layerSize=strlen(example->noteSequence);
  for(i=0;i<layerSize;i++)
    setActivation(inputLayer[i],
		  (example->noteSequence[i] > 72 ?
		   -1 :
		   example->noteSequence[i]/72));
    
  /*
   * Now back propagate, calculate and apply the deltas
   */
  scaleRating=(int)(1.27*example->rating);

  if(train) {

    for(i=0;i<7;i++) {
      sumSqrErr+=square(backProp(outputLayer[i],
				 ( (scaleRating & (0x1<<i)) >> i ) ));
                                                 /* back prop & get error */
      decimalValue+=(int)(rint(findOutput(outputLayer[i])))<<i;
    }
    
    for(i=0;i<7;i++)
      calcDeltas(outputLayer[i]);
                            /* calculate change in weights for this input */
    

    for(i=0;i<7;i++)
      applyDeltas(outputLayer[i]); /* change the weights */
      
    decimalError=fabs(example->rating-(double)(decimalValue)/1.27);
			     
  } else {
    for(i=0;i<7;i++) {
      sumSqrErr+=square( ( (scaleRating & (0x1<<i)) >> i )
			-findOutput(outputLayer[i]));
      decimalValue+=(int)(rint(findOutput(outputLayer[i])))<<i;
    }

    decimalError=fabs(example->rating-(double)(decimalValue)/1.27);
    
  }
  
  sumSqrErr/=7;
  
  /*
   * Now destroy the network, layer by layer
   */

  /* input layer */
  layerSize=strlen(example->noteSequence);
  for(i=0;i<layerSize;i++)
    freeNeuron(inputLayer[i]);
  free(inputLayer);

  /* first layer */
  layerSize=max(1,ceil(1+(layerSize-L_ONE_SPREAD)/L_ONE_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(firstLayer[i]);
  free(firstLayer);

  /* second layer */
  layerSize=max(1,ceil(1+(layerSize-L_TWO_SPREAD)/L_TWO_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(secondLayer[i]);
  free(secondLayer);
  
  /* third layer */
  layerSize=max(1,ceil(1+(layerSize-L_THREE_SPREAD)/L_THREE_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(thirdLayer[i]);
  free(thirdLayer);

  /* output node */
  for(i=0;i<7;i++)
    freeNeuron(outputLayer[i]);

  /*
   * now return with sumSqrErr, decimal Error
   */
  *decError=decimalError;
  return sumSqrErr;
  
} /* end of trialNet7 */

/*
 * trialNet1 - creates a network for the noteSequence in 'example', and then
 *   applies it to the 'notesequence', returning the squared error between
 *   the nets output and the rating in 'example'.  If 'train' is 1, it also
 *   trains the net for the noteSequence and rating in 'example', thus
 *   adjusting the weights and biases. 'decError' is the RMS error of the
 *   difference between the output*100 and the rating.
 *
 *   trialNet1 constructs a network with 1 output node.  The output of this
 *   node (which are treated
 *   as a 7-bit binary number.  The rating (1-100) is scaled to this range
 *   for comparisons.
 */
double trialNet1(RatingPair *example, int train, double *decError) {

  Neuron **inputLayer;
  Neuron **firstLayer;
  Neuron **secondLayer;
  Neuron **thirdLayer;
  Neuron *output;

  double sumSqrErr=0;
  int decimalError;
  int prevLayerSize, layerSize;
  int i,j;
  
  /*
   * Create the network layers
   */

  /* input layer */
  layerSize=strlen(example->noteSequence);
  inputLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    inputLayer[i]=createNeuron(0);

  /* first layer */
  layerSize=max(1,ceil(1+(layerSize-L_ONE_SPREAD)/L_ONE_DELTA));
  firstLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    firstLayer[i]=createNeuron(&oneBias);

  /* second layer */
  layerSize=max(1,ceil(1+(layerSize-L_TWO_SPREAD)/L_TWO_DELTA));
  secondLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    secondLayer[i]=createNeuron(&twoBias);

  /* third layer */
  layerSize=max(1,ceil(1+(layerSize-L_THREE_SPREAD)/L_THREE_DELTA));
  thirdLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    thirdLayer[i]=createNeuron(&threeBias);

  /* output nodes */
  output=createNeuron(&(outputBiases[0]));

  /*
   * Now make connections between the layers
   */

  /* connections from inputs to first layer nodes */
  prevLayerSize=layerSize=strlen(example->noteSequence);
  layerSize=max(1,ceil(1+(layerSize-L_ONE_SPREAD)/L_ONE_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<L_ONE_SPREAD;j++)
      if(L_ONE_DELTA*i+j < prevLayerSize)
	connectNeurons(inputLayer[(int)L_ONE_DELTA*i+j], firstLayer[i],
		       &(oneWeights[j]));

  /* connections from first layer to second layer nodes */
  prevLayerSize=layerSize;
  layerSize=max(1,ceil(1+(prevLayerSize-L_TWO_SPREAD)/L_TWO_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<L_TWO_SPREAD;j++)
      if(L_TWO_DELTA*i+j < prevLayerSize)
	connectNeurons(firstLayer[(int)L_TWO_DELTA*i+j], secondLayer[i],
		       &(twoWeights[j]));

  /* connections from second layer to third layer nodes */
  prevLayerSize=layerSize;
  layerSize=max(1,ceil(1+(prevLayerSize-L_THREE_SPREAD)/L_THREE_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<L_THREE_SPREAD;j++)
      if(L_THREE_DELTA*i+j < prevLayerSize)
	connectNeurons(secondLayer[(int)L_THREE_DELTA*i+j], thirdLayer[i],
		       &(threeWeights[j]));

  /* connections from third layer nodes to output node */
  for(i=0;i<layerSize;i++)
      connectNeurons(thirdLayer[i], output,
		     &(outputWeights[i]));

  /*
   * Now load the noteSequence values into the inputs
   */
  layerSize=strlen(example->noteSequence);
  for(i=0;i<layerSize;i++)
    setActivation(inputLayer[i],
		  (example->noteSequence[i] > 72 ?
		   -1 :
		   example->noteSequence[i]/72));
    
  /*
   * Now back propagate, calculate and apply the deltas
   */
  if(train) {

    sumSqrErr+=square(backProp(output, example->rating/100));
                                      /* back prop & get error */
    
    calcDeltas(output);  /* calculate change in weights for this input */
    
    applyDeltas(output); /* change the weights */
      
  } else
    sumSqrErr+=square( example->rating/100 -findOutput(output));
  
  decimalError=fabs(example->rating-100*findOutput(output));

  /*
   * Now destroy the network, layer by layer
   */

  /* input layer */
  layerSize=strlen(example->noteSequence);
  for(i=0;i<layerSize;i++)
    freeNeuron(inputLayer[i]);
  free(inputLayer);

  /* first layer */
  layerSize=max(1,ceil(1+(layerSize-L_ONE_SPREAD)/L_ONE_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(firstLayer[i]);
  free(firstLayer);

  /* second layer */
  layerSize=max(1,ceil(1+(layerSize-L_TWO_SPREAD)/L_TWO_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(secondLayer[i]);
  free(secondLayer);
  
  /* third layer */
  layerSize=max(1,ceil(1+(layerSize-L_THREE_SPREAD)/L_THREE_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(thirdLayer[i]);
  free(thirdLayer);

  /* output node */
  freeNeuron(output);

  /*
   * now return with sumSqrErr, decimal Error
   */
  *decError=decimalError;
  return sumSqrErr;
  
} /* end of trialNet1 */

int main(int argc, char *argv[]) {

  int i, j;
  int iterations=0;
  
  double tMaxSumSqrErr;    /* errors for training set */
  double tAvgSumSqrErr;
  double tMinSumSqrErr;
  double tAvgDecimalError;
  double tMaxDecimalError;
  double tMinDecimalError;

  int *tLengthBuckets;
  double *tErrorBuckets;

  int tDecErrorDistrib[100];
  
  double sumSqrErr;
  double decErr;
  
  double cMaxSumSqrErr;    /* errors for control set */
  double cAvgSumSqrErr;
  double cMinSumSqrErr;
  double cAvgDecimalError;
  double cMaxDecimalError;
  double cMinDecimalError;

  int *cLengthBuckets;
  double *cErrorBuckets;

  int cDecErrorDistrib[100];
  
  double maxErr=0.02;
  int maxIter=100;
  int stopType=AVG;
  double startTime;
  double lrate=0.5;
  double momentum=0;
  long int seedVal;
  int netType=ONE_OUTPUT;
  int checkFreq=100;
  char *statFileName=NULL;
  FILE *statFile=NULL;
  
  char *trainingSetName=NULL;
  int trainingSetSize;
  RatingPair *trainingSet;

  char *controlSetName=NULL;
  int controlSetSize;
  RatingPair *controlSet;

  int keepGoing=1; /* stop criteria */
  int different;
  int numSame=100;

  double *lastErrs;
  int lastErrPointer=0;

  /* make up a seedVal */
  seedVal=(long int)ReadSysClock();
  
  /* init with command line params */
  for(i=1;i<argc;i++) {
    
    if(!strcmp("-maxErr",argv[i])) {
      maxErr=atof(argv[++i]);
      stopType=MAX;
    } else if (!strcmp("-avgErr",argv[i])) {
      maxErr=atof(argv[++i]);
      stopType=AVG;
    } else if (!strcmp("-iter",argv[i])) {
      maxIter=atoi(argv[++i]);
      maxErr=0;
      stopType=ITER;
    } else if (!strcmp("-numSame",argv[i])) {
      numSame=atoi(argv[++i]);
    } else if(!strcmp("-lrate",argv[i])) {
      lrate=atof(argv[++i]);
    } else if(!strcmp("-momentum",argv[i])) {
      momentum=atof(argv[++i]);
    } else if(!strcmp("-seed",argv[i])) {
      seedVal=atoi(argv[++i]);
    } else if(!strcmp("-freq",argv[i])) {
      checkFreq=atoi(argv[++i]);
    } else if(!strcmp("-one",argv[i])) {
      netType=ONE_OUTPUT;
    } else if(!strcmp("-seven",argv[i])) {
      netType=SEVEN_OUTPUT;
    } else if(!strcmp("-o",argv[i])) {
      statFileName=argv[++i];
    } else if(!strcmp("-1spread",argv[i])) {
      L_ONE_SPREAD=atof(argv[++i]);
     } else if(!strcmp("-1delta",argv[i])) {
      L_ONE_DELTA=atof(argv[++i]);
    } else if(!strcmp("-2spread",argv[i])) {
      L_TWO_SPREAD=atof(argv[++i]);
    } else if(!strcmp("-2delta",argv[i])) {
      L_TWO_DELTA=atof(argv[++i]);
    } else if(!strcmp("-3spread",argv[i])) {
      L_THREE_SPREAD=atof(argv[++i]);
    } else if(!strcmp("-3delta",argv[i])) {
      L_THREE_DELTA=atof(argv[++i]);
    } else if(!strcmp("-h",argv[i])) {
      printf("\nSyntax: %s [-A] [-maxErr <val>] [-avgErr <val>] "
	     "[-iter <val>]\n\t\t[-numSame <val>] [-lrate <val>] "
	     "[-momentum <val>] "
	     "[-seed <val>] "
	     "\n\t\t[-one] [-seven] "
	     "[-1spread <val>] [-1delta <val>]\n\t\t[-2spread <val>] "
	     "[-2delta <val>] [-3spread <val>]\n\t\t[-3delta <val>] "
	     "[-maxLength <val>] [-freq <val>]\n\t\t[-o <Stat File>] "
	     "<Training File> <Control File>\n\n"
	     "\t-maxErr\t\tspecifies the target maximum error value\n"
	     "\t-avgErr\t\tspecifies the target average error value\n"
	     "\t-iter\t\tspecifies the iteration to stop at\n"
	     "\t-numSame\tspecifies how many identical errors to stop "
	     "after\n"
	     "\t-lrate\t\tspecifies the learning rate\n"
	     "\t-momentum\tspecifies the momentum for learning\n"
	     "\t-seed\t\tspecifies the random number generator seed\n"
	     "\t-one\t\tuse one output node network\n"
	     "\t-seven\t\tuse seven output node network\n"
	     "\t-1spread\tspecify spread of level one nodes\n"
	     "\t-1delta\t\tspecify level one node field shifts\n"
	     "\t-2spread\tspecify spread of level two nodes\n"
	     "\t-2delta\t\tspecify level one node field shifts\n"
	     "\t-3spread\tspecify spread of level one nodes\n"
	     "\t-3delta\t\tspecify level one node field shifts\n"
	     "\t-o\t\toutput statistics to the given file\n"
	     "\t-freq\t\tfrequency in training cycles to output stats\n\n"
	     "\tTraining File\tfile with training set\n"
	     "\tControl File\tfile with control set\n\n", argv[0]);
      exit(1);
    } else {
      if(trainingSetName==NULL) trainingSetName=argv[i];
      else controlSetName=argv[i];
    }
    
  }

  printf("\nTraining List Auto Rater with:\n\t%s=%4.3g\tlrate=%4.3g\t"
	 "momentum=%4.3g\tseed=%4.3ld\n\tUsing %s output node network.\n\n",
	 (stopType==AVG ? "avgErr" : "maxErr"), maxErr, lrate,
	 momentum, seedVal, (netType==SEVEN_OUTPUT ? "seven" : "one"));
  setLearnParams(lrate, momentum, lrate, momentum);

  /* allocate the lastErrs array */
  lastErrs=(double *)malloc(numSame*sizeof(double));
  for(i=0;i<numSame;i++) lastErrs[i]=i;
  
  /* create and load the training set array */
  trainingSetSize=loadDataSet(trainingSetName, &trainingSet);
  for(i=0;i<trainingSetSize;i++)
    L_MAX_LENGTH=max(L_MAX_LENGTH, strlen(trainingSet[i].noteSequence));
 
  /* create and load the control set array */
  controlSetSize=loadDataSet(controlSetName, &controlSet);
  for(i=0;i<controlSetSize;i++)
    L_MAX_LENGTH=max(L_MAX_LENGTH, strlen(controlSet[i].noteSequence));

  printf("Maximum Note Sequence Length=%g\n\n", L_MAX_LENGTH);

  L_MAX_LENGTH++;
  
  /* allocate and fill the length buffers */
  tErrorBuckets=(double *)malloc(L_MAX_LENGTH*sizeof(double));
  tLengthBuckets=(int *)malloc(L_MAX_LENGTH*sizeof(int));
  cErrorBuckets=(double *)malloc(L_MAX_LENGTH*sizeof(double));
  cLengthBuckets=(int *)malloc(L_MAX_LENGTH*sizeof(int));

  for(i=0;i<L_MAX_LENGTH;i++) {
    tLengthBuckets[i]=0;
    cLengthBuckets[i]=0;
  }

  for(i=0;i<trainingSetSize;i++)
    tLengthBuckets[strlen(trainingSet[i].noteSequence)]++;

  for(i=0;i<controlSetSize;i++) 
    cLengthBuckets[strlen(controlSet[i].noteSequence)]++;

  /* open stat file */
  if(statFileName!=NULL) {

    printf("Writing statistics to file '%s'\n\n", statFileName);

    statFile=fopen(statFileName, "w+");

    /* write out the run configuration */
    fwrite(&stopType, sizeof(int), 1, statFile);
    fwrite(&maxErr, sizeof(double), 1, statFile);
    fwrite(&lrate, sizeof(double), 1, statFile);
    fwrite(&seedVal, sizeof(long int), 1, statFile);
    fwrite(&netType, sizeof(int), 1, statFile);
    fwrite(&L_ONE_SPREAD, sizeof(double), 1, statFile);
    fwrite(&L_ONE_DELTA, sizeof(double), 1, statFile);
    fwrite(&L_TWO_SPREAD, sizeof(double), 1, statFile);
    fwrite(&L_TWO_DELTA, sizeof(double), 1, statFile);
    fwrite(&L_THREE_SPREAD, sizeof(double), 1, statFile);
    fwrite(&L_THREE_DELTA, sizeof(double), 1, statFile);
    fwrite(&L_MAX_LENGTH, sizeof(double), 1, statFile);
    fwrite(&checkFreq, sizeof(int), 1, statFile);
    fwrite(&trainingSetSize, sizeof(int), 1, statFile);
    fwrite(&controlSetSize, sizeof(int), 1, statFile);
    
    fwrite(tLengthBuckets, sizeof(int), L_MAX_LENGTH, statFile);
    fwrite(cLengthBuckets, sizeof(int), L_MAX_LENGTH, statFile);

  }
    
  /* initialize the random number generator */
  srand48(seedVal);
  
  /* initialize weights and biases */
  initNetVals();

  /* Now train the network until error falls below maxErr */
  startTime=ReadSysClock();
  do {

    /* initialize bucket arrays */
    for(i=0;i<L_MAX_LENGTH;i++) {
      tErrorBuckets[i]=0;
      cErrorBuckets[i]=0;
    }

    for(i=0;i<100;i++) {
      tDecErrorDistrib[i]=0;
      cDecErrorDistrib[i]=0;
    }
    
    /* reset errors */
    tAvgSumSqrErr=0;
    tMaxSumSqrErr=0;
    tMinSumSqrErr=1000;
    tAvgDecimalError=0;
    tMaxDecimalError=0;
    tMinDecimalError=1000;
    
    cAvgSumSqrErr=0;
    cMaxSumSqrErr=0;
    cMinSumSqrErr=1000;
    cAvgDecimalError=0;
    cMaxDecimalError=0;
    cMinDecimalError=0;
    
    /* for each sequence in training set, build a net and train the weights */
    for(i=0;i<trainingSetSize;i++) {
      if(netType==SEVEN_OUTPUT)
	sumSqrErr=trialNet7(&(trainingSet[i]), 1, &decErr);
      else
	sumSqrErr=trialNet1(&(trainingSet[i]), 1, &decErr);
      tMaxSumSqrErr=max(tMaxSumSqrErr, sumSqrErr);
      tMinSumSqrErr=min(tMinSumSqrErr, sumSqrErr);
      tAvgSumSqrErr+=sumSqrErr;
      tMaxDecimalError=max(tMaxDecimalError, decErr);
      tMinDecimalError=min(tMinDecimalError, decErr);
      tAvgDecimalError+=decErr;
      tErrorBuckets[strlen(trainingSet[i].noteSequence)]+=decErr;

      for(j=99;j>=0;j--)
	if(decErr > (double)j) { tDecErrorDistrib[j]++; break; }

    }
    
    tAvgSumSqrErr/=(double)trainingSetSize;
    tAvgDecimalError/=(double)trainingSetSize;

    /* track last numSame avgSumSqrErr errors as a stop condition */
    lastErrs[(lastErrPointer++)%numSame]=tAvgSumSqrErr;
      
    /* for each sequence in control set, build a net and get the results */
    for(i=0;i<controlSetSize;i++) {
      if(netType==SEVEN_OUTPUT)
	sumSqrErr=trialNet7(&(controlSet[i]), 0, &decErr);
      else
	sumSqrErr=trialNet1(&(controlSet[i]), 0, &decErr);	
      cMaxSumSqrErr=max(cMaxSumSqrErr, sumSqrErr);
      cMinSumSqrErr=min(cMinSumSqrErr, sumSqrErr);
      cAvgSumSqrErr+=sumSqrErr;
      cMaxDecimalError=max(cMaxDecimalError, decErr);
      cMinDecimalError=min(cMinDecimalError, decErr);
      cAvgDecimalError+=decErr;
      cErrorBuckets[strlen(controlSet[i].noteSequence)]+=decErr;

      for(j=99;j>=0;j--)
	if(decErr > (double)j) { cDecErrorDistrib[j]++; break; }

    }
    
    cAvgSumSqrErr/=(double)controlSetSize;
    cAvgDecimalError/=(double)controlSetSize;
    
    if(!(iterations%checkFreq)) {
      printf("Iteration %d:\t\n"
	     "  Training Set:\n"
	     "\tSum Sqr Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n"
	     "\tDecimal Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n"
	     "  Control Set:\n"
	     "\tSum Sqr Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n"
	     "\tDecimal Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n",
	     iterations, tAvgSumSqrErr, tMaxSumSqrErr, tMinSumSqrErr,
	     tAvgDecimalError, tMaxDecimalError, tMinDecimalError,
	     cAvgSumSqrErr, cMaxSumSqrErr, cMinSumSqrErr,
	     cAvgDecimalError, cMaxDecimalError, cMinDecimalError);
      fflush(stdout);

      /* if there is a stat file, write to it */
      if(statFileName!=NULL) {

	/* iteration number */
	fwrite(&iterations, sizeof(int), 1, statFile);
	
	/* training set stats */
	fwrite(&tAvgSumSqrErr, sizeof(double), 1, statFile);
	fwrite(&tMinSumSqrErr, sizeof(double), 1, statFile);
	fwrite(&tMaxSumSqrErr, sizeof(double), 1, statFile);
	fwrite(&tAvgDecimalError, sizeof(double), 1, statFile);
	fwrite(&tMinDecimalError, sizeof(double), 1, statFile);
	fwrite(&tMaxDecimalError, sizeof(double), 1, statFile);
	fwrite(tErrorBuckets, sizeof(double), L_MAX_LENGTH, statFile);
	fwrite(tDecErrorDistrib, sizeof(int), 100, statFile);
	
	/* control set stats */
	fwrite(&cAvgSumSqrErr, sizeof(double), 1, statFile);
	fwrite(&cMinSumSqrErr, sizeof(double), 1, statFile);
	fwrite(&cMaxSumSqrErr, sizeof(double), 1, statFile);
	fwrite(&cAvgDecimalError, sizeof(double), 1, statFile);
	fwrite(&cMinDecimalError, sizeof(double), 1, statFile);
	fwrite(&cMaxDecimalError, sizeof(double), 1, statFile);
	fwrite(cErrorBuckets, sizeof(double), L_MAX_LENGTH, statFile);
	fwrite(cDecErrorDistrib, sizeof(int), 100, statFile);
	
	/* now write out the weights and biases */
	fwriteNetState(statFile);

	/* flush to the file */
	fflush(statFile);
	
      }
    }
    
    iterations++;

    /* figure out stop criteria */
    if(stopType==ITER) {
      if(iterations==maxIter) keepGoing=0;
    } else {
      if(stopType==AVG) {
	if(tAvgSumSqrErr<maxErr) keepGoing=0;
      } else
	if(tMaxSumSqrErr<maxErr) keepGoing=0;
      
      /* if last numSame errors the same, exit */
      for(j=0,different=0;j<numSame;j++)
	if(lastErrs[0]!=lastErrs[j]) different++;
      if(!different) keepGoing=0;
    }  

  } while(keepGoing);

  /* print run statistics */
  printf("\nRun took %g seconds\n\n", ReadSysClock() - startTime);
  printf("Finished in %d Iterations.  Final Stats:\n"
	     "  Training Set:\n"
	     "\tSum Sqr Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n"
	     "\tDecimal Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n"
	     "  Control Set:\n"
	     "\tSum Sqr Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n"
	     "\tDecimal Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n",
	     iterations, tAvgSumSqrErr, tMaxSumSqrErr, tMinSumSqrErr,
	     tAvgDecimalError, tMaxDecimalError, tMinDecimalError,
	     cAvgSumSqrErr, cMaxSumSqrErr, cMinSumSqrErr,
	     cAvgDecimalError, cMaxDecimalError, cMinDecimalError);

  /* if there is a stat file and this is not a normal check point,
     write to it */
  if(statFileName!=NULL && iterations%checkFreq) {

    /* iteration number */
    fwrite(&iterations, sizeof(int), 1, statFile);
    
    /* training set stats */
    fwrite(&tAvgSumSqrErr, sizeof(double), 1, statFile);
    fwrite(&tMinSumSqrErr, sizeof(double), 1, statFile);
    fwrite(&tMaxSumSqrErr, sizeof(double), 1, statFile);
    fwrite(&tAvgDecimalError, sizeof(double), 1, statFile);
    fwrite(&tMinDecimalError, sizeof(double), 1, statFile);
    fwrite(&tMaxDecimalError, sizeof(double), 1, statFile);
    fwrite(tErrorBuckets, sizeof(double), L_MAX_LENGTH, statFile);
    fwrite(tDecErrorDistrib, sizeof(int), 100, statFile);
    
    /* control set stats */
    fwrite(&cAvgSumSqrErr, sizeof(double), 1, statFile);
    fwrite(&cMinSumSqrErr, sizeof(double), 1, statFile);
    fwrite(&cMaxSumSqrErr, sizeof(double), 1, statFile);
    fwrite(&cAvgDecimalError, sizeof(double), 1, statFile);
    fwrite(&cMinDecimalError, sizeof(double), 1, statFile);
    fwrite(&cMaxDecimalError, sizeof(double), 1, statFile);
    fwrite(cErrorBuckets, sizeof(double), L_MAX_LENGTH, statFile);
    fwrite(cDecErrorDistrib, sizeof(int), 100, statFile);
    
    /* now write out the weights and biases */
    fwriteNetState(statFile);
    
    /* flush to the file */
    fflush(statFile);
    
  }
      
  /* free buffers and close files */
  for(i=0;i<controlSetSize;i++)
    free(controlSet[i].noteSequence);
  free(controlSet);
  for(i=0;i<trainingSetSize;i++)
    free(trainingSet[i].noteSequence);
  free(trainingSet);
  
  free(tErrorBuckets);
  free(tLengthBuckets);
  free(cErrorBuckets);
  free(cLengthBuckets);

  free(oneWeights);
  free(twoWeights);
  free(threeWeights);
  free(outputWeights);

  free(lastErrs);
  
  fclose(statFile);
  
  return 0;
  
}

