#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define max(xVal,yVal) ( (xVal) > (yVal) ? (xVal) : (yVal) )

/* stop criterion types */
#define AVG 0
#define MAX 1
#define ITER 2

/* network configurations */
#define ONE_OUTPUT 0
#define SEVEN_OUTPUT 1

/* different plot data that can be generated */
#define NONE 0
#define T_SUM_ERRORS 1
#define C_SUM_ERRORS 2
#define T_VS_C_SUM_ERRORS 3
#define T_DEC_ERRORS 4
#define C_DEC_ERRORS 5
#define T_VS_C_DEC_ERRORS 6
#define T_LENGTH_HISTOGRAM 7
#define C_LENGTH_HISTOGRAM 8
#define T_ERROR_DISTRIB 9
#define C_ERROR_DISTRIB 10

/* the network weight arrays */
double *oneWeights;
double oneBias;
double *twoWeights;
double twoBias;
double *threeWeights;
double threeBias;
double *outputWeights;
double outputBiases[7];		  

/*
 * Defines the level spreads for the list auto rater
 *
 *   The spread of the top level integrator will be adjusted to accept
 *   the maximum number of inputs from level 3 given the maximimum sequence
 *   length is L_MAX_LENGTH.
 */
double L_ONE_SPREAD=8;
double L_ONE_DELTA=4;
double L_TWO_SPREAD=8;
double L_TWO_DELTA=4;
double L_THREE_SPREAD=4;
double L_THREE_DELTA=2;
double L_MAX_LENGTH=0;

/*
 * initNetVals - allocates the weight arrays.
 */
void initNetVals() {

  double outputLength;
  
  /* create the weight and bias arrays */
  oneWeights=(double *)malloc(L_ONE_SPREAD*sizeof(double));
  twoWeights=(double *)malloc(L_TWO_SPREAD*sizeof(double));
  threeWeights=(double *)malloc(L_THREE_SPREAD*sizeof(double));

  outputLength=max(1,ceil(1+(L_MAX_LENGTH-L_ONE_SPREAD)/L_ONE_DELTA));
  outputLength=max(1,ceil(1+(outputLength-L_TWO_SPREAD)/L_TWO_DELTA));
  outputLength=max(1,ceil(1+(outputLength-L_THREE_SPREAD)/L_THREE_DELTA));

  outputLength*=7;
  
  outputWeights=(double *)malloc(outputLength*sizeof(double));

} /* end of initNetVals */


/*
 * freadNetState - reads all of the networks weights and biases from
 *   the given file into the internal arrays.  If writeWeights is true,
 *   the weights and parameters are written to a file, "list-weights.data"
 */
void freadNetState(FILE *inputFile, int writeWeights) {

  double outputLength;
  FILE *weightFile;
  
  fread(oneWeights, sizeof(double), L_ONE_SPREAD, inputFile);
  fread(&oneBias, sizeof(double), 1, inputFile);

  fread(twoWeights, sizeof(double), L_TWO_SPREAD, inputFile);
  fread(&twoBias, sizeof(double), 1, inputFile);

  fread(threeWeights, sizeof(double), L_THREE_SPREAD, inputFile);
  fread(&threeBias, sizeof(double), 1, inputFile);

  outputLength=max(1,ceil(1+(L_MAX_LENGTH-L_ONE_SPREAD)/L_ONE_DELTA));
  outputLength=max(1,ceil(1+(outputLength-L_TWO_SPREAD)/L_TWO_DELTA));
  outputLength=max(1,ceil(1+(outputLength-L_THREE_SPREAD)/L_THREE_DELTA));

  outputLength*=7;
    
  fread(outputWeights, sizeof(double), (size_t)outputLength, inputFile);
  fread(outputBiases, sizeof(double), 7, inputFile);

  /* dump the weights and other info if requested */
  if(writeWeights) {

    /* open the file to write to */
    weightFile=fopen("list-weights.data","w+");
    
    /* write out the basic parameters */
    fwrite(&L_ONE_SPREAD, sizeof(double), 1, weightFile);
    fwrite(&L_ONE_DELTA, sizeof(double), 1, weightFile);
    fwrite(&L_TWO_SPREAD, sizeof(double), 1, weightFile);
    fwrite(&L_TWO_DELTA, sizeof(double), 1, weightFile);
    fwrite(&L_THREE_SPREAD, sizeof(double), 1, weightFile);
    fwrite(&L_THREE_DELTA, sizeof(double), 1, weightFile);
    fwrite(&L_MAX_LENGTH, sizeof(double), 1, weightFile);
    
    /* write the weights and biases */
    fwrite(oneWeights, sizeof(double), L_ONE_SPREAD, weightFile);
    fwrite(&oneBias, sizeof(double), 1, weightFile);
    
    fwrite(twoWeights, sizeof(double), L_TWO_SPREAD, weightFile);
    fwrite(&twoBias, sizeof(double), 1, weightFile);
    
    fwrite(threeWeights, sizeof(double), L_THREE_SPREAD, weightFile);
    fwrite(&threeBias, sizeof(double), 1, weightFile);
    
    fwrite(outputWeights, sizeof(double), (size_t)outputLength, weightFile);
    fwrite(outputBiases, sizeof(double), 7, weightFile);

    /* close the file to write to */
    fclose(weightFile);
    
  }

} /* end of freadNetState */

int main(int argc, char *argv[]) {

  int i;
  int iterations=0;
  
  double tMaxSumSqrErr;    /* errors for training set */
  double tAvgSumSqrErr;
  double tMinSumSqrErr;
  double tAvgDecimalError;
  double tMaxDecimalError;
  double tMinDecimalError;

  int *tLengthBuckets;
  double *tErrorBuckets;

  int tDecErrorDistrib[100];
  
  double cMaxSumSqrErr;    /* errors for control set */
  double cAvgSumSqrErr;
  double cMinSumSqrErr;
  double cAvgDecimalError;
  double cMaxDecimalError;
  double cMinDecimalError;

  int *cLengthBuckets;
  double *cErrorBuckets;

  int cDecErrorDistrib[100];
  
  double maxErr=0.02;
  int stopType=AVG;
  double lrate=0.5;
  double momentum=0;
  long int seedVal;
  int netType=ONE_OUTPUT;
  int checkFreq=100;
  char *statFileName=NULL;
  FILE *statFile=NULL;

  int trainingSetSize;
  int controlSetSize;

  int alphaType=0;
  int plotType=NONE;
  int plotIteration=0;
  int quietMode=0;

  FILE *dataFile=NULL;
  char *dataFileName=NULL;

  int plotMade=0;

  int writeWeights=1;
  
  /* init with command line params */
  for(i=1;i<argc;i++) {
    
    if(!strcmp("-alpha",argv[i])) {
      alphaType=1;
    } else if(!strcmp("-tSumErr",argv[i])) {
      plotType=T_SUM_ERRORS;
    } else if(!strcmp("-cSumErr",argv[i])) {
      plotType=C_SUM_ERRORS;
    } else if(!strcmp("-tcSumErr",argv[i])) {
      plotType=T_VS_C_SUM_ERRORS;
    } else if(!strcmp("-tDecErr",argv[i])) {
      plotType=T_DEC_ERRORS;
    } else if(!strcmp("-cDecErr",argv[i])) {
      plotType=C_DEC_ERRORS;
    } else if(!strcmp("-tcDecErr",argv[i])) {
      plotType=T_VS_C_DEC_ERRORS;
    } else if(!strcmp("-tLengthHist",argv[i])) {
      plotType=T_LENGTH_HISTOGRAM;
      plotIteration=atoi(argv[++i]);
    } else if(!strcmp("-cLengthHist",argv[i])) {
      plotType=C_LENGTH_HISTOGRAM;
      plotIteration=atoi(argv[++i]);
    } else if(!strcmp("-tErrorDist",argv[i])) {
      plotType=T_ERROR_DISTRIB;
      plotIteration=atoi(argv[++i]);
    } else if(!strcmp("-cErrorDist",argv[i])) {
      plotType=C_ERROR_DISTRIB;
      plotIteration=atoi(argv[++i]);
    } else if(!strcmp("-o",argv[i])) {
      dataFileName=argv[++i];
    } else if(!strcmp("-q",argv[i])) {
      quietMode=1;
    } else if(!strcmp("-h",argv[i])) {
      printf("\nSyntax: %s [-alpha] [-tSumErr|-cSumErr|-tcSumErr|-tDecErr"
	     "|-cDecErr|\n\t\t-tcDecErr|-tLengthHist <iteration>|-cLengthHist "
	     "<iteration>|\n\t\t"
	     "-tErrorDist <iteration>|-cErrorDist <iteration>]\n\t\t"
	     "[-q] [-o <Data File>] <Statistics File>\n\n"
	     "\t-alpha\t\t\tstat file generated on DEC Alpha\n"
	     "\t-tSumErr\t\tgenerate sum square error data for training set\n"
	     "\t-cSumErr\t\tgenerate sum square error data for control set\n"
	     "\t-tcSumErr\t\tgenerate training versus control set sum square"
	     "\n\t\t\t\terror data\n"
	     "\t-tDecErr\t\tgenerate decimal error data for training set\n"
	     "\t-cDecErr\t\tgenerate decimal error data for control set\n"
	     "\t-tcDecErr\t\tgenerate training versus control set decimal"
	     "\n\t\t\t\terror data\n"
	     "\t-tLengthHist\t\tgenerate training set decimal error vs "
	     "length\n\t\t\t\thistogram for the given iteration\n"
	     "\t-cLengthHist\t\tgenerate control set decimal error vs "
	     "length\n\t\t\t\thistogram for the given iteration\n"
	     "\t-tErrorDist\t\tgenerate training set decimal error "
	     "distribution\n\t\t\t\tdata for the given iteration\n"
	     "\t-cErrorDist\t\tgenerate control set decimal error "
	     "distribution\n\t\t\t\tdata for the given iteration\n"
	     "\t-q\t\t\tsupresses information for each iteration\n"
	     "\t-o\t\t\toutput data to specified data file\n\n"
	     "\tStatistics File\t\tstat file output by l-auto-rater\n",
	     argv[0]);
      exit(1);
    } else {
      statFileName=argv[i];
    }
    
  }

  /* open the statistics file */
  printf("\nReading statistics file '%s'\n\n", statFileName);
  
  statFile=fopen(statFileName, "r");
  
  /* print the plot type being used */
  if(plotType==NONE)
    printf("No output data being generated.\n\n");
  else
    printf("Generating data for a ");
  
  switch(plotType) {

  case NONE: break;
  case T_SUM_ERRORS: printf("training set sum error"); break;
  case C_SUM_ERRORS: printf("control set sum error"); break;
  case T_VS_C_SUM_ERRORS: printf("training and control set sum error"); break;
  case T_DEC_ERRORS: printf("training set decimal error"); break;
  case C_DEC_ERRORS: printf("control set decimal error"); break;
  case T_VS_C_DEC_ERRORS: printf("training and control set decimal error");
    break;
  case T_LENGTH_HISTOGRAM: printf("training set decimal error vs. length");
    break;
  case C_LENGTH_HISTOGRAM: printf("control set decimal error vs. length");
    break;
  case T_ERROR_DISTRIB: printf("training set decimal error distribution");
    break;
  case C_ERROR_DISTRIB: printf("control set decimal error distribution");
    break;

  default: printf("impossible");

  }
  
  if(plotType!=NONE)
    printf(" plot.\n\n");
  
  /* open output file */
  if(plotType!=NONE)
    if(dataFileName==NULL) {
      dataFile=fopen("output.data", "w+");
      printf("Outputting data to file 'output.data'.\n\n");
    } else {
      dataFile=fopen(dataFileName, "w+");
      printf("Outputting data to file '%s'.\n\n", dataFileName);
    }
  
  /* read in the run configuration */
  fread(&stopType, sizeof(int), 1, statFile);
  fread(&maxErr, sizeof(double), 1, statFile);
  fread(&lrate, sizeof(double), 1, statFile);
  fread(&seedVal, sizeof(long int), 1, statFile);
  if(alphaType) fseek(statFile, 4, SEEK_CUR);
  fread(&netType, sizeof(int), 1, statFile);
  fread(&L_ONE_SPREAD, sizeof(double), 1, statFile);
  fread(&L_ONE_DELTA, sizeof(double), 1, statFile);
  fread(&L_TWO_SPREAD, sizeof(double), 1, statFile);
  fread(&L_TWO_DELTA, sizeof(double), 1, statFile);
  fread(&L_THREE_SPREAD, sizeof(double), 1, statFile);
  fread(&L_THREE_DELTA, sizeof(double), 1, statFile);
  fread(&L_MAX_LENGTH, sizeof(double), 1, statFile);
  fread(&checkFreq, sizeof(int), 1, statFile);
  fread(&trainingSetSize, sizeof(int), 1, statFile);
  fread(&controlSetSize, sizeof(int), 1, statFile);

  printf("Training set size is %d, control set %d.\n\n",
	 trainingSetSize, controlSetSize);
  
  /* allocate arrays */
  tErrorBuckets=(double *)malloc(L_MAX_LENGTH*sizeof(double));
  tLengthBuckets=(int *)malloc(L_MAX_LENGTH*sizeof(int));
  cErrorBuckets=(double *)malloc(L_MAX_LENGTH*sizeof(double));
  cLengthBuckets=(int *)malloc(L_MAX_LENGTH*sizeof(int));

  /* read in distribution of lengths for control and training sets */
  fread(tLengthBuckets, sizeof(int), L_MAX_LENGTH, statFile);
  fread(cLengthBuckets, sizeof(int), L_MAX_LENGTH, statFile);

  /* print run configuration */
  printf("List Auto Rater Run trained with:\n\t%s=%4.3g\tlrate=%4.3g\t"
	 "momentum=%4.3g\tseed=%4.3ld\n\tUsing %s output node network.\n\n",
	 (stopType==AVG ? "avgErr" : "maxErr"), maxErr, lrate,
	 momentum, seedVal, (netType==SEVEN_OUTPUT ? "seven" : "one"));
  
  /* initialize weights and biases arrays */
  initNetVals();

  /* Now read all stats until EOF */
  do {

    /*
     * read info from the stat file
     */
    
    /* iteration number */
    fread(&iterations, sizeof(int), 1, statFile);
    
    /* training set stats */
    fread(&tAvgSumSqrErr, sizeof(double), 1, statFile);
    fread(&tMinSumSqrErr, sizeof(double), 1, statFile);
    fread(&tMaxSumSqrErr, sizeof(double), 1, statFile);
    fread(&tAvgDecimalError, sizeof(double), 1, statFile);
    fread(&tMinDecimalError, sizeof(double), 1, statFile);
    fread(&tMaxDecimalError, sizeof(double), 1, statFile);
    fread(tErrorBuckets, sizeof(double), L_MAX_LENGTH, statFile);
    fread(tDecErrorDistrib, sizeof(int), 100, statFile);
    
    /* control set stats */
    fread(&cAvgSumSqrErr, sizeof(double), 1, statFile);
    fread(&cMinSumSqrErr, sizeof(double), 1, statFile);
    fread(&cMaxSumSqrErr, sizeof(double), 1, statFile);
    fread(&cAvgDecimalError, sizeof(double), 1, statFile);
    fread(&cMinDecimalError, sizeof(double), 1, statFile);
    fread(&cMaxDecimalError, sizeof(double), 1, statFile);
    fread(cErrorBuckets, sizeof(double), L_MAX_LENGTH, statFile);
    fread(cDecErrorDistrib, sizeof(int), 100, statFile);
    
    /* now read in the weights and biases */
    if(iterations==plotIteration && !plotMade && writeWeights)
      freadNetState(statFile, 1);
    else
      freadNetState(statFile, 0);
    
    /* print iteration info */
    if(!quietMode) 
      printf("Iteration %d:\t\n"
	     "  Training Set:\n"
	     "\tSum Sqr Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n"
	     "\tDecimal Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n"
	     "  Control Set:\n"
	     "\tSum Sqr Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n"
	     "\tDecimal Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n\n",
	     iterations, tAvgSumSqrErr, tMaxSumSqrErr, tMinSumSqrErr,
	     tAvgDecimalError, tMaxDecimalError, tMinDecimalError,
	     cAvgSumSqrErr, cMaxSumSqrErr, cMinSumSqrErr,
	     cAvgDecimalError, cMaxDecimalError, cMinDecimalError);
    
    /* write to appropriate output files */
    switch(plotType) {
      
    case NONE: break;

    case T_SUM_ERRORS:
      fprintf(dataFile, "%d\t%g\t%g\t%g\n", iterations, tMaxSumSqrErr,
	      tAvgSumSqrErr, tMinSumSqrErr);
      break;

    case C_SUM_ERRORS:
      fprintf(dataFile, "%d\t%g\t%g\t%g\n", iterations, cMaxSumSqrErr,
	      cAvgSumSqrErr, cMinSumSqrErr);
      break;

    case T_VS_C_SUM_ERRORS:
      fprintf(dataFile, "%d\t%g\t%g\n", iterations, tAvgSumSqrErr,
	      cAvgSumSqrErr);
      break;

    case T_DEC_ERRORS:
      fprintf(dataFile, "%d\t%g\t%g\t%g\n", iterations, tMaxDecimalError,
	      tAvgDecimalError, tMinDecimalError);
      break;

    case C_DEC_ERRORS:
      fprintf(dataFile, "%d\t%g\t%g\t%g\n", iterations, cMaxDecimalError,
	      cAvgDecimalError, cMinDecimalError);
      break;

    case T_VS_C_DEC_ERRORS:
      fprintf(dataFile, "%d\t%g\t%g\n", iterations, tAvgDecimalError,
	      cAvgDecimalError);
      break;

    case T_LENGTH_HISTOGRAM:
      if(iterations==plotIteration && !plotMade++)
	for(i=0;i<L_MAX_LENGTH;i++)
	  if(tLengthBuckets[i]!=0)
	    fprintf(dataFile, "%d\t%g\n", i,
		    tErrorBuckets[i]/(double)tLengthBuckets[i]);
	  else
	    fprintf(dataFile, "%d\t0\n", i);
      break;

    case C_LENGTH_HISTOGRAM:
      if(iterations==plotIteration && !plotMade++)
	for(i=0;i<L_MAX_LENGTH;i++)
	  if(cLengthBuckets[i]!=0)
	    fprintf(dataFile, "%d\t%g\n", i,
		    cErrorBuckets[i]/(double)cLengthBuckets[i]);
	  else
	    fprintf(dataFile, "%d\t0\n", i);
      break;

    case T_ERROR_DISTRIB:
      if(iterations==plotIteration && !plotMade++)
	for(i=0;i<100;i++)
	  fprintf(dataFile, "%d\t%g\n", i,
		  (double)tDecErrorDistrib[i]/(double)trainingSetSize);   
      break;

    case C_ERROR_DISTRIB:
      if(iterations==plotIteration && !plotMade++)
	for(i=0;i<100;i++)
	  fprintf(dataFile, "%d\t%g\n", i,
		  (double)cDecErrorDistrib[i]/(double)controlSetSize);   
      break;
      
    default: printf("Unknown plot type!\n\n");

    }

  } while(!feof(statFile));

  /* print how many iterations in the file */
  printf("Processed %d learning cycles.\n\n", iterations);
  
  /* free arrays */
  free(tErrorBuckets);
  free(tLengthBuckets);
  free(cErrorBuckets);
  free(cLengthBuckets);

  free(oneWeights);
  free(twoWeights);
  free(threeWeights);
  free(outputWeights);

  /* close files */
  fclose(statFile);
  fclose(dataFile);
  
  return 0;
  
}



