#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "neural.h"

/* these two macros define the activation function and its derivative */
#define func(xVal) 1/(1+exp(-(xVal)));
#define dfunc(xVal) exp(-(xVal))/((1+exp(-(xVal)))*(1+exp(-(xVal))))

/* these store parameters for the backpropagation learning */
double wlrate=0.5;     /* learning rate for weights */
double wmomentum=0;  /* momentum for weights */

double blrate=0.5;     /* learning rate for weights */
double bmomentum=0;  /* momentum for weights */

typedef struct connection {

  double *weight;

  Neuron *Input;
  Neuron *Output;

  double deltaW;        /* the previous weight change delta */
  double errWeightDeriv;     /* the current weight change delta */
  int wDeltaCalced;    /* indicates if the weight has been changed this round
			   (this prevents double changing during recursion) */
  
} Connection;

struct neuron {

  double *bias;

  Connection **Inputs;
  int numInputs;
  Connection **Outputs;
  int numOutputs;
  
  double activation;
  int valid;            /* 1 if activation is valid (has been updated) */

  double de_da_i;
  double slope_net;     /* f'(netInput) */
  
  double deltaB;        /* the previous bias change delta */
  double errBiasDeriv;     /* the current bias change delta */
  int bDeltaCalced;      /* indicates if the weight has been changed this round
			   (this prevents double changing during recursion) */
  
};

/*
 * setLearnParams - sets the learning rate and momentum for both weights
 *   (w_lrate and w_momentum) and biases (b_lrate and b_momentum)
 */
void setLearnParams(double w_lrate, double w_momentum,
		    double b_lrate, double b_momentum)
{

  wlrate=w_lrate;
  wmomentum=w_momentum;
  blrate=b_lrate;
  bmomentum=b_momentum;
  
} /* end of setLearnParams */

/*
 * createNeuron - returns a new allocated and initialiazed neuron, with its
 *   bias stored in the location pointed to by bias
 */
Neuron *createNeuron(double *bias) {

  Neuron *newNeuron;

  newNeuron=(Neuron *)malloc(sizeof(Neuron));

  newNeuron->bias=bias;

  newNeuron->Inputs=NULL;
  newNeuron->numInputs=0;
  newNeuron->Outputs=NULL;
  newNeuron->numOutputs=0;

  newNeuron->activation=0;
  newNeuron->valid=0;

  newNeuron->de_da_i=0;
  newNeuron->slope_net=0;

  newNeuron->bDeltaCalced=0;
  newNeuron->deltaB=0;
  newNeuron->errBiasDeriv=0;
  
  return newNeuron;
  
} /* end of createNeuron */

/*
 * backPropRecurse - determines and propagates backwards the error fom the
 *   given output neuron, using its activation and the given partialErrAct,
 *   which is a component of de_da_i for this neuron.  In the case of this
 *   being an input neuron, it should be: t_i - a_i, where t_i is the target
 *   for this neuron, and a_i is its current activation.
 */
void backPropRecurse(Neuron *aNeuron, double partialErrAct) {

  int i;
  double partialErrNet; /* the component of the net derivative error
			   from this part of the activation derivative
			   error */
  
  aNeuron->de_da_i+=partialErrAct;

  partialErrNet=partialErrAct*aNeuron->slope_net;
  
  for(i=0;i<aNeuron->numInputs;i++) 
    backPropRecurse(aNeuron->Inputs[i]->Input,
		    *(aNeuron->Inputs[i]->weight) * partialErrNet);
  
} /* end of backPropRecurse */

/*
 * backProp - when called for each output neuron with its desired value,
 *   will back propagate all error through the net.  It will not calculate
 *   new weights.  For that, 'adjustNetwork' or 'calcDeltas' and 'adjustDeltas'
 *   must be called.
 *
 *   This function may be called at any time after the input node values
 *   have been set.  It should not be called more than once before
 *   'resetNet'.
 *
 *   It returns the current error for this output.
 */
double backProp(Neuron *outputNeuron, double targetValue) {

  if(!(outputNeuron->valid))
    findOutput(outputNeuron);
  
  backPropRecurse(outputNeuron, targetValue - outputNeuron->activation);

  return targetValue - outputNeuron->activation;
	 
} /* end of backProp */

/*
 * adjustNetwork - when called for each output node, will adjust all
 *   weights and biases in the network based on the previous back
 *   propagation phase.  This must be called after backProp has been called
 *   for each ouput node in order to work properly.  It should not be called
 *   more than once before a 'resetNet'
 */
void adjustNetwork(Neuron *outputNeuron) {

  calcDeltas(outputNeuron);
  applyDeltas(outputNeuron);
  
} /* end of adjustNetwork */

/*
 * calcDeltas - when called for each output node, will determine all
 *   deltas for weights and biases in the network based on the previous back
 *   propagation phase, and add them to the currently accumulated delta
 *   (from previous training examples).  This must be called after backProp
 *   has been called for a specific desired output in order to work properly.
 *   It should not be called more than once before a 'resetNet'.  To actually
 *   apply the deltas, 'applyDeltas' must be called.
 */
void calcDeltas(Neuron *outputNeuron) {

  int i;
  Connection *aConnection;
  double errNet;

  /* calculate the derivative of error wrt net input */
  errNet=outputNeuron->de_da_i * outputNeuron->slope_net;

  /* calculate bias delta for this neuron, if not already calculated */
  if(!(outputNeuron->bDeltaCalced)) {
    outputNeuron->errBiasDeriv+=errNet;
    outputNeuron->bDeltaCalced=1;
  }
  
  /* now calculate deltas for input connections and neurons */
  for(i=0;i<outputNeuron->numInputs;i++) {

    aConnection=outputNeuron->Inputs[i];
    
    /* calculate connection delta, if not already calculated */
    if(!(aConnection->wDeltaCalced)) {
      aConnection->errWeightDeriv+=errNet*aConnection->Input->activation;
      aConnection->wDeltaCalced=1;
    }
    
    /* recursively calculate deltas */
    calcDeltas(aConnection->Input);
    
  }
  
} /* end of calcDeltas */

/*
 * applyDeltas - when called for each output node, will apply the accumulated
 *   delta to change the weights and biases in the network.  The accumulated
 *   delta is reset, and moved to the old delta value for use in the momentum
 *   term for future delta calculations.  This should not be called more than
 *   once before a new backProp and calcDelta phase.
 */
void applyDeltas(Neuron *outputNeuron) {

  int i;
  Connection *aConnection;
  double delta;

  /* apply the new Delta to the bias for this neuron, if it is not 0 */
  if(outputNeuron->errBiasDeriv) {
    delta=blrate*outputNeuron->errBiasDeriv+bmomentum*outputNeuron->deltaB;
    *(outputNeuron->bias)+=delta;
    outputNeuron->deltaB=delta;
    outputNeuron->errBiasDeriv=0;
  }
  
  /* now adjust input connections and neurons */
  for(i=0;i<outputNeuron->numInputs;i++) {

    aConnection=outputNeuron->Inputs[i];
    
    /* adjust connection, if not already adjusted */
    if(aConnection->errWeightDeriv) {
      delta=wlrate*aConnection->errWeightDeriv+wmomentum*aConnection->deltaW;
      *(aConnection->weight)+=delta;
      aConnection->deltaW=delta;
      aConnection->errWeightDeriv=0;
    }
    
    /* recursively adjust neurons */
    applyDeltas(aConnection->Input);
    
  }
  
} /* end of applyDeltas */

/*
 * findOutput - determines the current activation level of the given
 *   neuron, assuming the activation of the input neurons have been set
 *   to the appropriate inputs.  It will recurse back through all higher
 *   layers of neurons.
 */
double findOutput(Neuron *aNeuron) {

  double netInput;
  int i;

  /* first see if our current activation is valid */
  if(aNeuron->valid)
    return aNeuron->activation;
  
  /* if there are no inputs from other neurons, must be an input neuron,
     but activation wasn't valid, so there must be an error */
  if(aNeuron->Inputs==0) {
    fprintf(stderr, "Invalid input neuron found: Exiting...\n\n");
    exit(1);
  }

  /* otherwise, sum (weights x outputs) for all input connections */
  for(i=0,netInput=0;i<aNeuron->numInputs;i++)
    netInput+=findOutput(aNeuron->Inputs[i]->Input)*
      (*(aNeuron->Inputs[i]->weight));

  /* add the bias term */
  netInput+= *(aNeuron->bias);

  /* run it through the sigmoid function */
  aNeuron->activation=func(netInput);
  aNeuron->slope_net=dfunc(netInput);
    
  /* set the result as valid for the future */
  aNeuron->valid=1;

  return aNeuron->activation;
    
} /* end of findOutput */

/*
 * resetNet - when called in turn for each output neuron, it will reset
 *   the network and prepare it for evaluation with a new set of inputs.
 *   It must be called to clear cached activation levels in the neurons.
 */
void resetNet(Neuron *outputNeuron) {

  int i;
  
  /* reset ourself */
  outputNeuron->valid=0;
  outputNeuron->bDeltaCalced=0;
  outputNeuron->de_da_i=0;
  outputNeuron->slope_net=0;
  
  /* reset all input connections to us */
  for(i=0;i<outputNeuron->numInputs;i++)
    outputNeuron->Inputs[i]->wDeltaCalced=0;


  /* reset our input neurons */
  if(outputNeuron->numInputs!=0)
    for(i=0;i<outputNeuron->numInputs;i++)
      resetNet(outputNeuron->Inputs[i]->Input);

  return;
  
} /* end of resetNet */


/*
 * connectNeurons - makes a connection between two neurons.  Weight is a
 *   pointer to the location to store the weight (potentially with initial
 *   value)
 */
void connectNeurons(Neuron *inputNeuron, Neuron *outputNeuron, \
		    double *weight)
{

  Connection *newConnection;

  /* create and initialize new connection */
  newConnection=(Connection *)malloc(sizeof(Connection));

  newConnection->weight=weight;
  newConnection->Input=inputNeuron;
  newConnection->Output=outputNeuron;
  newConnection->wDeltaCalced=0;
  newConnection->deltaW=0;
  newConnection->errWeightDeriv=0;
  
  /* attach the connection to the inputNeuron */
  if(inputNeuron->numOutputs == 0)
    inputNeuron->Outputs=(Connection **)
      malloc(++(inputNeuron->numOutputs)*sizeof(Connection *));
  else
    inputNeuron->Outputs=(Connection **)
      realloc(inputNeuron->Outputs,
	      ++(inputNeuron->numOutputs)*sizeof(Connection *));

  inputNeuron->Outputs[inputNeuron->numOutputs-1]=newConnection;
  
  /* attach the connection to the outputNeuron */
  if(outputNeuron->numInputs == 0)
    outputNeuron->Inputs=(Connection **)
      malloc(++(outputNeuron->numInputs)*sizeof(Connection *));
  else
    outputNeuron->Inputs=(Connection **)
      realloc(outputNeuron->Inputs,
	      ++(outputNeuron->numInputs)*sizeof(Connection *));

  outputNeuron->Inputs[outputNeuron->numInputs-1]=newConnection;

  return;
  
} /* end of connectNeurons */

/*
 * setActivation - sets the activation of the given neuron (presumably an
 *   input neuron)
 */
void setActivation(Neuron *inputNeuron, double activation) {

  inputNeuron->activation=activation;
  inputNeuron->valid=1;
  inputNeuron->bDeltaCalced=1; /* otherwise crash on bias adjust for inputs */
  
  return;
  
} /* end of setActivation */

/*
 * freeNeuron - frees a neuron and all its related data structures
 */
void freeNeuron(Neuron *aNeuron) {

  int i;
  
  /*
   * free all the input connections.  Note that by doing this,
   * all connections are freed once and only once if freeNeuron is called
   * for each neuron in the net
   */
  for(i=0;i<aNeuron->numInputs;i++)
    free(aNeuron->Inputs[i]);

  /* free the Input and Output arrays */
  free(aNeuron->Inputs);
  free(aNeuron->Outputs);

  /* free the neuron itself */
  free(aNeuron);
  
} /* end of freeNeuron */








