
#ifndef _NEURAL_H_

#define _NEURAL_H_

typedef struct neuron Neuron;

/*
 * setLearnParams - sets the learning rate and momentum for both weights
 *   (w_lrate and w_momentum) and biases (b_lrate and b_momentum)
 */
void setLearnParams(double w_lrate, double w_momentum,
		    double b_lrate, double b_momentum);

/*
 * createNeuron - returns a new allocated and initialiazed neuron, with its
 *   bias stored in the location pointed to by bias
 */
Neuron *createNeuron(double *bias);  

/*
 * findOutput - determines the current activation level of the given
 *   neuron, assuming the activation of the input neurons have been set
 *   to the appropriate inputs.  It will recurse back through all higher
 *   layers of neurons.
 */
double findOutput(Neuron *aNeuron);

/*
 * backProp - when called for each output neuron with its desired value,
 *   will back propagate all error through the net.  It will not calculate
 *   new weights.  For that, 'adjustNetwork' or 'calcDeltas' and 'adjustDeltas'
 *   must be called.
 *
 *   This function may be called at any time after the input node values
 *   have been set.  It should not be called more than once before
 *   'resetNet'.
 *
 *   It returns the current error for this output.
 */
double backProp(Neuron *outputNeuron, double targetValue);

/*
 * adjustNetwork - when called for each output node, will adjust all
 *   weights and biases in the network based on the previous back
 *   propagation phase.  This must be called after backProp has been called
 *   for each ouput node in order to work properly.  It should not be called
 *   more than once before a 'resetNet'
 */
void adjustNetwork(Neuron *outputNeuron);

/*
 * calcDeltas - when called for each output node, will determine all
 *   deltas for weights and biases in the network based on the previous back
 *   propagation phase, and add them to the currently accumulated delta
 *   (from previous training examples).  This must be called after backProp
 *   has been called for a specific desired output in order to work properly.
 *   It should not be called more than once before a 'resetNet'.  To actually
 *   apply the deltas, 'applyDeltas' must be called.
 */
void calcDeltas(Neuron *outputNeuron);

/*
 * applyDeltas - when called for each output node, will apply the accumulated
 *   delta to change the weights and biases in the network.  The accumulated
 *   delta is reset, and moved to the old delta value for use in the momentum
 *   term for future delta calculations.  This should not be called more than
 *   once before a new backProp and calcDelta phase.
 */
void applyDeltas(Neuron *outputNeuron);

/*
 * resetNet - when called in turn for each output neuron, it will reset
 *   the network and prepare it for evaluation with a new set of inputs.
 *   It must be called to clear cached activation levels in the neurons.
 */
void resetNet(Neuron *outputNeuron);

/*
 * connectNeurons - makes a connection between two neurons.  Weight is a
 *   pointer to the location to store the weight (potentially with initial
 *   value)
 */
void connectNeurons(Neuron *inputNeuron, Neuron *outputNeuron, \
		    double *weight);

/*
 * setActivation - sets the activation of the given neuron (presumably an
 *   input neuron)
 */
void setActivation(Neuron *inputNeuron, double activation);

/*
 * freeNeuron - frees a neuron and all its related data structures, including
 *   connection information
 */
void freeNeuron(Neuron *aNeuron);

#endif








