#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include "neural.h"

#define square(xVal) (xVal)*(xVal)
#define max(xVal,yVal) ( (xVal) > (yVal) ? (xVal) : (yVal) )
#define min(xVal,yVal) ( (xVal) < (yVal) ? (xVal) : (yVal) )

/* stop criterion types */
#define AVG 0
#define MAX 1
#define ITER 2

/* network comparison typ */
#define BINARY 0
#define COMPLEX 1

typedef struct {

  int ind1;
  int ind2;
  int result; /* -2 -> ind1 << ind2, etc. */
  
} Comparison;

double *aOneWeights;
double aOneBias;
double *aTwoWeights;
double aTwoBias;
double *aThreeWeights;
double aThreeBias;
double *aOutputWeights;

double *bOneWeights;
double bOneBias;
double *bTwoWeights;
double bTwoBias;
double *bThreeWeights;
double bThreeBias;
double *bOutputWeights;

double outputBias;		  

/*
 * Defines the level spreads for the tournament auto rater
 *
 *   The spread of the top level integrator will be adjusted to accept
 *   the maximum number of inputs from level 3 given the maximimum sequence
 *   length is T_MAX_LENGTH.
 */
double T_ONE_SPREAD=8;
double T_ONE_DELTA=4;
double T_TWO_SPREAD=8;
double T_TWO_DELTA=4;
double T_THREE_SPREAD=4;
double T_THREE_DELTA=2;
double T_MAX_LENGTH=0;

/*
 * ReadSysClock - returns the time in seconds and fractions of seconds
 *     as a double (for use only for calculating deltas)
 */
double
ReadSysClock ()
{

    struct timeval tv;

    (void) gettimeofday(&tv, NULL);
    return (tv.tv_sec +tv.tv_usec / 1000000.0);

} /* end of ReadSysClock */

/*
 * initNetVals - allocates the weight arrays, and initializes weighs and
 *   biases to small random values
 */
void initNetVals() {

  double outputLength;
  int i;
  
  /* create the weight and bias arrays */
  aOneWeights=(double *)malloc(T_ONE_SPREAD*sizeof(double));
  aTwoWeights=(double *)malloc(T_TWO_SPREAD*sizeof(double));
  aThreeWeights=(double *)malloc(T_THREE_SPREAD*sizeof(double));

  bOneWeights=(double *)malloc(T_ONE_SPREAD*sizeof(double));
  bTwoWeights=(double *)malloc(T_TWO_SPREAD*sizeof(double));
  bThreeWeights=(double *)malloc(T_THREE_SPREAD*sizeof(double));

  outputLength=max(1,ceil(1+(T_MAX_LENGTH-T_ONE_SPREAD)/T_ONE_DELTA));
  outputLength=max(1,ceil(1+(outputLength-T_TWO_SPREAD)/T_TWO_DELTA));
  outputLength=max(1,ceil(1+(outputLength-T_THREE_SPREAD)/T_THREE_DELTA));

  aOutputWeights=(double *)malloc(outputLength*sizeof(double));
  bOutputWeights=(double *)malloc(outputLength*sizeof(double));

  /* fill the weight arrays and biases with random values */
  for(i=0;i<T_ONE_SPREAD;i++) {
    aOneWeights[i]=( drand48() > 0.5 ? 1 : -1)*drand48();
    bOneWeights[i]=( drand48() > 0.5 ? 1 : -1)*drand48();
  }
  for(i=0;i<T_TWO_SPREAD;i++) {
    aTwoWeights[i]=( drand48() > 0.5 ? 1 : -1)*drand48();
    bTwoWeights[i]=( drand48() > 0.5 ? 1 : -1)*drand48();
  }
  for(i=0;i<T_THREE_SPREAD;i++) {
    aThreeWeights[i]=( drand48() > 0.5 ? 1 : -1)*drand48();
    bThreeWeights[i]=( drand48() > 0.5 ? 1 : -1)*drand48();
  }
  for(i=0;i<outputLength;i++) {
    aOutputWeights[i]=( drand48() > 0.5 ? 1 : -1)*drand48();
    bOutputWeights[i]=( drand48() > 0.5 ? 1 : -1)*drand48();
  }
  
  aOneBias=( drand48() > 0.5 ? 1 : -1)*drand48();
  bOneBias=( drand48() > 0.5 ? 1 : -1)*drand48();
  aTwoBias=( drand48() > 0.5 ? 1 : -1)*drand48();
  bTwoBias=( drand48() > 0.5 ? 1 : -1)*drand48();
  aThreeBias=( drand48() > 0.5 ? 1 : -1)*drand48();
  bThreeBias=( drand48() > 0.5 ? 1 : -1)*drand48();

  outputBias=( drand48() > 0.5 ? 1 : -1)*drand48();

} /* end of initNetVals */

/*
 * fwriteNetState - writes all of the networks weights and biases into
 *   the given file
 */
void fwriteNetState(FILE *outputFile) {

  double outputLength;
  
  fwrite(aOneWeights, sizeof(double), T_ONE_SPREAD, outputFile);
  fwrite(&aOneBias, sizeof(double), 1, outputFile);

  fwrite(bOneWeights, sizeof(double), T_ONE_SPREAD, outputFile);
  fwrite(&bOneBias, sizeof(double), 1, outputFile);

  fwrite(aTwoWeights, sizeof(double), T_TWO_SPREAD, outputFile);
  fwrite(&aTwoBias, sizeof(double), 1, outputFile);

  fwrite(bTwoWeights, sizeof(double), T_TWO_SPREAD, outputFile);
  fwrite(&bTwoBias, sizeof(double), 1, outputFile);

  fwrite(aThreeWeights, sizeof(double), T_THREE_SPREAD, outputFile);
  fwrite(&aThreeBias, sizeof(double), 1, outputFile);

  fwrite(bThreeWeights, sizeof(double), T_THREE_SPREAD, outputFile);
  fwrite(&bThreeBias, sizeof(double), 1, outputFile);

  outputLength=max(1,ceil(1+(T_MAX_LENGTH-T_ONE_SPREAD)/T_ONE_DELTA));
  outputLength=max(1,ceil(1+(outputLength-T_TWO_SPREAD)/T_TWO_DELTA));
  outputLength=max(1,ceil(1+(outputLength-T_THREE_SPREAD)/T_THREE_DELTA));

  fwrite(aOutputWeights, sizeof(double), (size_t)outputLength, outputFile);
  fwrite(bOutputWeights, sizeof(double), (size_t)outputLength, outputFile);
  fwrite(&outputBias, sizeof(double), 1, outputFile);

} /* end of fwriteNetState */

/*
 * loadDataSet - loads the noteSequences and comparisons in the file named
 *   'setFileName'.  The comparisons go into the array 'theSet', while the
 *   sequences go into 'theSequences'.  Both are allocated to fit
 *   the number of individuals and comparisons in the file.  The last element
 *   of the set array has ind1 set to '-1' and indicates the end of the
 *   array.  The last element of 'theSequences' is set to NULL.
 *   The function returns the number of comparisons in the set.
 */
int loadDataSet(char *setFileName, Comparison **theSet, char ***theSequences) {

  int i;
  int numComps;
  int numInds;
  int temp;
  FILE *setFile;
  Comparison *newArray;
  char **newSeqArray;
  
  /* open the file */
  setFile=fopen(setFileName, "r");

  /* find out how many comparisons in the set */
  fread(&numComps, sizeof(int), 1, setFile);

  /* allocate the comparison array */
  newArray=(Comparison *)malloc((numComps+1)*sizeof(Comparison));
  
  /* now loop through loading comparisons */
  for(i=0;i<numComps;i++) 
    fread(&(newArray[i]), sizeof(Comparison), 1, setFile);

  /* set the end of comparison array */
  newArray[i].ind1=-1;

  /* find out how many sequences */
  fread(&numInds, sizeof(int), 1, setFile);

  /* allocate the sequence array */
  newSeqArray=(char **)malloc((numInds+1)*sizeof(char *));
  
  /* now loop through loading the sequences */
  for(i=0;i<numInds;i++) {

    /* read size of sequence */
    fread(&temp, sizeof(int), 1, setFile);
      
    /* allocate and read note sequence */
    newSeqArray[i]=(char *)malloc(temp*sizeof(char));
    fread(newSeqArray[i], sizeof(char), temp, setFile);

  }

  /* set the end of the sequence array */
  newSeqArray[i]=NULL;
  
  /* close the file */
  fclose(setFile);
  
  /* set the array pointers for return */
  *theSet=newArray;
  *theSequences=newSeqArray;
  
  return numComps;

} /* end of loadDataSet */

/*
 * trialNet - creates a network for the noteSequences referred to by
 *   'example', and then applies it to the two sequences, returning the
 *   squared error between the nets output and the comparison result.  The
 *   nets output is taken to be ind1 < ind2 if output closer to 0, and
 *   inds 1 > ind2 if output closer to 1.  If 'train' is 1, it also trains
 *   the net for the comparison in 'example', thus adjusting the weights and
 *   biases.
 *
 *   The individuals in the comparison are taken as offsets to noteSequences
 *   in 'theSequences'
 *
 *   If 'outputMode' is 'BINARY', the net trains to just indicate which
 *   sequence is better.  If 'COMPLEX', it is trained for much worse, worse,
 *   equal, better, and much better.
 *
 *   'correct' is set to 1 if the net outputs the correct relationship
 *   between the two sequences.
 */
double trialNet(Comparison *example, char **theSequences, int train,
		int outputMode, int *correct) {

  Neuron **aInputLayer;
  Neuron **bInputLayer;
  Neuron **aFirstLayer;
  Neuron **bFirstLayer;
  Neuron **aSecondLayer;
  Neuron **bSecondLayer;
  Neuron **aThirdLayer;
  Neuron **bThirdLayer;
  Neuron *output;

  double sumSqrErr=0;
  double desiredOutput;
  int prevLayerSize, layerSize;
  int i,j;
  
  /*
   * Create the 'A' network layers (for individual 1)
   */
  
  /* input layer */
  layerSize=strlen(theSequences[example->ind1]);
  aInputLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    aInputLayer[i]=createNeuron(0);

  /* first layer */
  layerSize=max(1,ceil(1+(layerSize-T_ONE_SPREAD)/T_ONE_DELTA));
  aFirstLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    aFirstLayer[i]=createNeuron(&aOneBias);

  /* second layer */
  layerSize=max(1,ceil(1+(layerSize-T_TWO_SPREAD)/T_TWO_DELTA));
  aSecondLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    aSecondLayer[i]=createNeuron(&aTwoBias);

  /* third layer */
  layerSize=max(1,ceil(1+(layerSize-T_THREE_SPREAD)/T_THREE_DELTA));
  aThirdLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    aThirdLayer[i]=createNeuron(&aThreeBias);

  /*
   * Create the 'B' network layers (for individual 2)
   */
  
  /* input layer */
  layerSize=strlen(theSequences[example->ind2]);
  bInputLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    bInputLayer[i]=createNeuron(0);

  /* first layer */
  layerSize=max(1,ceil(1+(layerSize-T_ONE_SPREAD)/T_ONE_DELTA));
  bFirstLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    bFirstLayer[i]=createNeuron(&bOneBias);

  /* second layer */
  layerSize=max(1,ceil(1+(layerSize-T_TWO_SPREAD)/T_TWO_DELTA));
  bSecondLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    bSecondLayer[i]=createNeuron(&bTwoBias);

  /* third layer */
  layerSize=max(1,ceil(1+(layerSize-T_THREE_SPREAD)/T_THREE_DELTA));
  bThirdLayer=(Neuron **)malloc(layerSize*sizeof(Neuron *));
  for(i=0;i<layerSize;i++)
    bThirdLayer[i]=createNeuron(&bThreeBias);

  /*
   * Create the output node
   */
  output=createNeuron(&outputBias);

  /*
   * Now make connections between the layers of side 'A'
   */

  /* connections from inputs to first layer nodes */
  prevLayerSize=layerSize=strlen(theSequences[example->ind1]);
  layerSize=max(1,ceil(1+(layerSize-T_ONE_SPREAD)/T_ONE_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<T_ONE_SPREAD;j++)
      if(T_ONE_DELTA*i+j < prevLayerSize)
	connectNeurons(aInputLayer[(int)T_ONE_DELTA*i+j], aFirstLayer[i],
		       &(aOneWeights[j]));

  /* connections from first layer to second layer nodes */
  prevLayerSize=layerSize;
  layerSize=max(1,ceil(1+(prevLayerSize-T_TWO_SPREAD)/T_TWO_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<T_TWO_SPREAD;j++)
      if(T_TWO_DELTA*i+j < prevLayerSize)
	connectNeurons(aFirstLayer[(int)T_TWO_DELTA*i+j], aSecondLayer[i],
		       &(aTwoWeights[j]));

  /* connections from second layer to third layer nodes */
  prevLayerSize=layerSize;
  layerSize=max(1,ceil(1+(prevLayerSize-T_THREE_SPREAD)/T_THREE_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<T_THREE_SPREAD;j++)
      if(T_THREE_DELTA*i+j < prevLayerSize)
	connectNeurons(aSecondLayer[(int)T_THREE_DELTA*i+j], aThirdLayer[i],
		       &(aThreeWeights[j]));

  /* connections from third layer nodes to output node */
  for(i=0;i<layerSize;i++)
      connectNeurons(aThirdLayer[i], output,
		     &(aOutputWeights[i]));

  /*
   * Now make connections between the layers of side 'B'
   */

  /* connections from inputs to first layer nodes */
  prevLayerSize=layerSize=strlen(theSequences[example->ind2]);
  layerSize=max(1,ceil(1+(layerSize-T_ONE_SPREAD)/T_ONE_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<T_ONE_SPREAD;j++)
      if(T_ONE_DELTA*i+j < prevLayerSize)
	connectNeurons(bInputLayer[(int)T_ONE_DELTA*i+j], bFirstLayer[i],
		       &(bOneWeights[j]));

  /* connections from first layer to second layer nodes */
  prevLayerSize=layerSize;
  layerSize=max(1,ceil(1+(prevLayerSize-T_TWO_SPREAD)/T_TWO_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<T_TWO_SPREAD;j++)
      if(T_TWO_DELTA*i+j < prevLayerSize)
	connectNeurons(bFirstLayer[(int)T_TWO_DELTA*i+j], bSecondLayer[i],
		       &(bTwoWeights[j]));

  /* connections from second layer to third layer nodes */
  prevLayerSize=layerSize;
  layerSize=max(1,ceil(1+(prevLayerSize-T_THREE_SPREAD)/T_THREE_DELTA));

  for(i=0;i<layerSize;i++)
    for(j=0;j<T_THREE_SPREAD;j++)
      if(T_THREE_DELTA*i+j < prevLayerSize)
	connectNeurons(bSecondLayer[(int)T_THREE_DELTA*i+j], bThirdLayer[i],
		       &(bThreeWeights[j]));

  /* connections from third layer nodes to output node */
  for(i=0;i<layerSize;i++)
      connectNeurons(bThirdLayer[i], output,
		     &(bOutputWeights[i]));

  /*
   * Now load the noteSequences values into the inputs
   */

  layerSize=strlen(theSequences[example->ind1]);
  for(i=0;i<layerSize;i++)
    setActivation(aInputLayer[i],
		  ((theSequences[example->ind1])[i] > 72 ?
		   -1 :
		   (theSequences[example->ind1])[i]/72));

  layerSize=strlen(theSequences[example->ind2]);
  for(i=0;i<layerSize;i++)
    setActivation(bInputLayer[i],
		  ((theSequences[example->ind2])[i] > 72 ?
		   -1 :
		   (theSequences[example->ind2])[i]/72));
    
  /*
   * Now back propagate, calculate and apply the deltas
   */
  desiredOutput=(example->result==4 ? 0 : example->result);
  
  if(outputMode==BINARY)
    desiredOutput=(desiredOutput<0 ? 0 : 1);
  else
    desiredOutput=(desiredOutput+2)/4;   
  
  if(train) {

    sumSqrErr+=square(backProp(output, desiredOutput));
                                      /* back prop & get error */
    
    calcDeltas(output);  /* calculate change in weights for this input */
    
    applyDeltas(output); /* change the weights */
      
  } else
    sumSqrErr+=square( desiredOutput - findOutput(output));

  desiredOutput=(example->result==4 ? 0 : example->result);
  desiredOutput=(desiredOutput<0 ? 0 : 1);
    
  if(fabs(desiredOutput-findOutput(output)) < 0.5)
    *correct=1;
  else
    *correct=0;
  
  /*
   * Now destroy the network, layer by layer
   */

  /*
   * Destroy side 'A'
   */
  
  /* input layer */
  layerSize=strlen(theSequences[example->ind1]);
  for(i=0;i<layerSize;i++)
    freeNeuron(aInputLayer[i]);
  free(aInputLayer);

  /* first layer */
  layerSize=max(1,ceil(1+(layerSize-T_ONE_SPREAD)/T_ONE_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(aFirstLayer[i]);
  free(aFirstLayer);

  /* second layer */
  layerSize=max(1,ceil(1+(layerSize-T_TWO_SPREAD)/T_TWO_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(aSecondLayer[i]);
  free(aSecondLayer);
  
  /* third layer */
  layerSize=max(1,ceil(1+(layerSize-T_THREE_SPREAD)/T_THREE_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(aThirdLayer[i]);
  free(aThirdLayer);

  /*
   * Destroy side 'B'
   */
  
  /* input layer */
  layerSize=strlen(theSequences[example->ind2]);
  for(i=0;i<layerSize;i++)
    freeNeuron(bInputLayer[i]);
  free(bInputLayer);

  /* first layer */
  layerSize=max(1,ceil(1+(layerSize-T_ONE_SPREAD)/T_ONE_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(bFirstLayer[i]);
  free(bFirstLayer);

  /* second layer */
  layerSize=max(1,ceil(1+(layerSize-T_TWO_SPREAD)/T_TWO_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(bSecondLayer[i]);
  free(bSecondLayer);
  
  /* third layer */
  layerSize=max(1,ceil(1+(layerSize-T_THREE_SPREAD)/T_THREE_DELTA));
  for(i=0;i<layerSize;i++)
    freeNeuron(bThirdLayer[i]);
  free(bThirdLayer);

  /*
   * free the output node
   */
  freeNeuron(output);

  /*
   * now return with sumSqrErr
   */
  return sumSqrErr;
  
} /* end of trialNet */

int main(int argc, char *argv[]) {

  int i, j;
  int iterations=0;

  double tMaxSumSqrErr;    /* errors for training set */
  double tAvgSumSqrErr;
  double tMinSumSqrErr;
  double tPercentCorrect;
  
  int *tLengthBuckets;
  double *tErrorBuckets;
  
  int tErrorDistrib[100];
    
  double sumSqrErr;
  int correct;
  
  double cMaxSumSqrErr;    /* errors for control set */
  double cAvgSumSqrErr;
  double cMinSumSqrErr;
  double cPercentCorrect;
  
  int *cLengthBuckets;
  double *cErrorBuckets;

  int cErrorDistrib[100];
  
  double maxErr=0.02;
  int maxIter=100;
  int stopType=AVG;
  int outputMode=BINARY;
  double startTime;
  double lrate=0.5;
  double momentum=0;
  long int seedVal;
  int checkFreq=100;
  char *statFileName=NULL;
  FILE *statFile=NULL;
  
  char *trainingSetName=NULL;
  int trainingSetSize;
  Comparison *trainingSet;
  char **trainingSetSequences;
  
  char *controlSetName=NULL;
  int controlSetSize;
  Comparison *controlSet;
  char **controlSetSequences;

  int keepGoing=1; /* stop criteria */
  int different;
  int numSame=100;

  double *lastErrs;
  int lastErrPointer=0;

  /* make up a seedVal */
  seedVal=(long int)ReadSysClock();
  
  /* init with command line params */
  for(i=1;i<argc;i++) {
    
    if(!strcmp("-maxErr",argv[i])) {
      maxErr=atof(argv[++i]);
      stopType=MAX;
    } else if (!strcmp("-avgErr",argv[i])) {
      maxErr=atof(argv[++i]);
      stopType=AVG;
    } else if (!strcmp("-iter",argv[i])) {
      maxIter=atoi(argv[++i]);
      maxErr=0;
      stopType=ITER;
    } else if (!strcmp("-numSame",argv[i])) {
      numSame=atoi(argv[++i]);
    } else if(!strcmp("-lrate",argv[i])) {
      lrate=atof(argv[++i]);
    } else if(!strcmp("-momentum",argv[i])) {
      momentum=atof(argv[++i]);
    } else if(!strcmp("-seed",argv[i])) {
      seedVal=atoi(argv[++i]);
    } else if(!strcmp("-freq",argv[i])) {
      checkFreq=atoi(argv[++i]);
    } else if(!strcmp("-binary",argv[i])) {
      outputMode=BINARY;
    } else if(!strcmp("-complex",argv[i])) {
      outputMode=COMPLEX;
    } else if(!strcmp("-o",argv[i])) {
      statFileName=argv[++i];
    } else if(!strcmp("-1spread",argv[i])) {
      T_ONE_SPREAD=atof(argv[++i]);
    } else if(!strcmp("-1delta",argv[i])) {
      T_ONE_DELTA=atof(argv[++i]);
    } else if(!strcmp("-2spread",argv[i])) {
      T_TWO_SPREAD=atof(argv[++i]);
    } else if(!strcmp("-2delta",argv[i])) {
      T_TWO_DELTA=atof(argv[++i]);
    } else if(!strcmp("-3spread",argv[i])) {
      T_THREE_SPREAD=atof(argv[++i]);
    } else if(!strcmp("-3delta",argv[i])) {
      T_THREE_DELTA=atof(argv[++i]);
    } else if(!strcmp("-h",argv[i])) {
      printf("\nSyntax: %s [-A] [-maxErr <val>] [-avgErr <val>] "
	     "[-iter <val>]\n\t\t[-numSame <val>] [-lrate <val>] "
	     "[-momentum <val>] [-seed <val>]\n"
	     "\t\t[-binary] [-complex] "
	     "[-1spread <val>] [-1delta <val>]\n\t\t[-2spread <val>] "
	     "[-2delta <val>] [-3spread <val>]\n\t\t[-3delta <val>] "
	     "[-maxLength <val>] [-freq <val>]\n\t\t[-o <Stat File>] "
	     "<Training File> <Control File>\n\n"
	     "\t-maxErr\t\tspecifies the target maximum error value\n"
	     "\t-avgErr\t\tspecifies the target average error value\n"
	     "\t-iter\t\tspecifies the iteration to stop at\n"
	     "\t-numSame\tspecifies how many identical errors to stop "
	     "after\n"
	     "\t-lrate\t\tspecifies the learning rate\n"
	     "\t-momentum\tspecifies the momentum for learning\n"
	     "\t-seed\t\tspecifies the random number generator seed\n"
	     "\t-binary\t\ttrain net for binary output\n"
	     "\t-complex\ttrain net for multi-level (complex) output\n"
	     "\t-1spread\tspecify spread of level one nodes\n"
	     "\t-1delta\t\tspecify level one node field shifts\n"
	     "\t-2spread\tspecify spread of level two nodes\n"
	     "\t-2delta\t\tspecify level one node field shifts\n"
	     "\t-3spread\tspecify spread of level one nodes\n"
	     "\t-3delta\t\tspecify level one node field shifts\n"
	     "\t-o\t\toutput statistics to the given file\n"
	     "\t-freq\t\tfrequency in training cycles to output stats\n\n"
	     "\tTraining File\tfile with training set\n"
	     "\tControl File\tfile with control set\n\n", argv[0]);
      exit(1);
    } else {
      if(trainingSetName==NULL) trainingSetName=argv[i];
      else controlSetName=argv[i];
    }
    
  }

  printf("\nTraining Tournament Auto Rater with:\n\t%s=%4.3g\tlrate=%4.3g\t"
	 "momentum=%4.3g\tseed=%4.3ld\n\tUsing %s network output mode.\n\n",
	 (stopType==AVG ? "avgErr" : "maxErr"), maxErr, lrate,
	 momentum, seedVal, (outputMode==BINARY ? "binary" : "complex"));
  setLearnParams(lrate, momentum, lrate, momentum);

  /* allocate the lastErrs array */
  lastErrs=(double *)malloc(numSame*sizeof(double));
  for(i=0;i<numSame;i++) lastErrs[i]=i;
  
  /* create and load the training set array */
  trainingSetSize=loadDataSet(trainingSetName, &trainingSet,
			      &trainingSetSequences);
  for(i=0;trainingSetSequences[i]!=NULL;i++)
    T_MAX_LENGTH=max(T_MAX_LENGTH, strlen(trainingSetSequences[i]));
 
  /* create and load the control set array */
  controlSetSize=loadDataSet(controlSetName, &controlSet,
			     &controlSetSequences);
  for(i=0;controlSetSequences[i]!=NULL;i++)
    T_MAX_LENGTH=max(T_MAX_LENGTH, strlen(controlSetSequences[i]));

  printf("Maximum Note Sequence Length=%g\n\n", T_MAX_LENGTH);

  T_MAX_LENGTH++;
  
  /* allocate and fill the length buffers */
  tErrorBuckets=(double *)malloc(T_MAX_LENGTH*T_MAX_LENGTH*sizeof(double));
  tLengthBuckets=(int *)malloc(T_MAX_LENGTH*T_MAX_LENGTH*sizeof(int));
  cErrorBuckets=(double *)malloc(T_MAX_LENGTH*T_MAX_LENGTH*sizeof(double));
  cLengthBuckets=(int *)malloc(T_MAX_LENGTH*T_MAX_LENGTH*sizeof(int));

  for(i=0;i<T_MAX_LENGTH*T_MAX_LENGTH;i++) {
    tLengthBuckets[i]=0;
    cLengthBuckets[i]=0;
  }

  for(i=0;i<trainingSetSize;i++)
    tLengthBuckets[(int)(T_MAX_LENGTH*
		  strlen(trainingSetSequences[trainingSet[i].ind1])+
		  strlen(trainingSetSequences[trainingSet[i].ind2]))]++;

  for(i=0;i<controlSetSize;i++)
    cLengthBuckets[(int)(T_MAX_LENGTH*
		  strlen(controlSetSequences[controlSet[i].ind1])+
		  strlen(controlSetSequences[controlSet[i].ind2]))]++;
  
  /* open stat file */
  if(statFileName!=NULL) {

    printf("Writing statistics to file '%s'\n\n", statFileName);

    statFile=fopen(statFileName, "w+");

    /* write out the run configuration */
    fwrite(&stopType, sizeof(int), 1, statFile);
    fwrite(&maxErr, sizeof(double), 1, statFile);
    fwrite(&lrate, sizeof(double), 1, statFile);
    fwrite(&seedVal, sizeof(long int), 1, statFile);
    fwrite(&outputMode, sizeof(int), 1, statFile);
    fwrite(&T_ONE_SPREAD, sizeof(double), 1, statFile);
    fwrite(&T_ONE_DELTA, sizeof(double), 1, statFile);
    fwrite(&T_TWO_SPREAD, sizeof(double), 1, statFile);
    fwrite(&T_TWO_DELTA, sizeof(double), 1, statFile);
    fwrite(&T_THREE_SPREAD, sizeof(double), 1, statFile);
    fwrite(&T_THREE_DELTA, sizeof(double), 1, statFile);
    fwrite(&T_MAX_LENGTH, sizeof(double), 1, statFile);
    fwrite(&checkFreq, sizeof(int), 1, statFile);
    fwrite(&trainingSetSize, sizeof(int), 1, statFile);
    fwrite(&controlSetSize, sizeof(int), 1, statFile);
    fwrite(tLengthBuckets, sizeof(int), T_MAX_LENGTH*T_MAX_LENGTH, statFile);
    fwrite(cLengthBuckets, sizeof(int), T_MAX_LENGTH*T_MAX_LENGTH, statFile);

  }
    
  /* initialize the random number generator */
  srand48(seedVal);
  
  /* initialize weights and biases */
  initNetVals();

  /* Now train the network until error falls below maxErr */
  startTime=ReadSysClock();
  do {

    /* initialize bucket arrays */
    for(i=0;i<T_MAX_LENGTH*T_MAX_LENGTH;i++) {
      tErrorBuckets[i]=0;
      cErrorBuckets[i]=0;
    }
    
    for(i=0;i<100;i++) {
      tErrorDistrib[i]=0;
      cErrorDistrib[i]=0;
    }
    
    /* reset errors */
    tAvgSumSqrErr=0;
    tMaxSumSqrErr=0;
    tMinSumSqrErr=1000;
    tPercentCorrect=0;
    
    cAvgSumSqrErr=0;
    cMaxSumSqrErr=0;
    cMinSumSqrErr=1000;
    cPercentCorrect=0;
    
    /* for each sequence in training set, build a net and train the weights */
    for(i=0;i<trainingSetSize;i++) {

      sumSqrErr=trialNet(&(trainingSet[i]), trainingSetSequences,
			 1, outputMode, &correct);
      
      tMaxSumSqrErr=max(tMaxSumSqrErr, sumSqrErr);
      tMinSumSqrErr=min(tMinSumSqrErr, sumSqrErr);
      tAvgSumSqrErr+=sumSqrErr;

      if(correct) tPercentCorrect++;
      
      tErrorBuckets[(int)(T_MAX_LENGTH*
		   strlen(trainingSetSequences[trainingSet[i].ind1])+
		   strlen(trainingSetSequences[trainingSet[i].ind2]))]
	+=sumSqrErr;

      for(j=99;j>=0;j--)
	if(sumSqrErr > (double)j/100.0) { tErrorDistrib[j]++; break; }

    }
    
    tAvgSumSqrErr/=(double)trainingSetSize;
    tPercentCorrect/=(double)trainingSetSize;

    /* track last numSame avgSumSqrErr errors as a stop condition */
    lastErrs[(lastErrPointer++)%numSame]=tAvgSumSqrErr;
    
    /* for each sequence in control set, build a net and get the results */
    for(i=0;i<controlSetSize;i++) {

      sumSqrErr=trialNet(&(controlSet[i]), controlSetSequences,
			 0, outputMode, &correct);

      cMaxSumSqrErr=max(cMaxSumSqrErr, sumSqrErr);
      cMinSumSqrErr=min(cMinSumSqrErr, sumSqrErr);
      cAvgSumSqrErr+=sumSqrErr;

      if(correct) cPercentCorrect++;
      
      cErrorBuckets[(int)(T_MAX_LENGTH*
		   strlen(controlSetSequences[controlSet[i].ind1])+
		   strlen(controlSetSequences[controlSet[i].ind2]))]
	+=sumSqrErr;

      for(j=99;j>=0;j--)
	if(sumSqrErr > (double)j/100.0) { cErrorDistrib[j]++; break; }

    }
    
    cAvgSumSqrErr/=(double)controlSetSize;
    cPercentCorrect/=(double)controlSetSize;
    
    if(!(iterations%checkFreq)) {
      printf("Iteration %d:\t\n"
	     "  Training Set:\n"
	     "\tSum Sqr Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n"
	     "\tPercent Correct:\t\t%4.3g\n"
	     "  Control Set:\n"
	     "\tSum Sqr Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n"
	     "\tPercent Correct:\t\t%4.3g\n\n",
	     iterations, tAvgSumSqrErr, tMaxSumSqrErr, tMinSumSqrErr,
	     tPercentCorrect*100, cAvgSumSqrErr, cMaxSumSqrErr, cMinSumSqrErr,
	     cPercentCorrect*100);
      fflush(stdout);

      /* if there is a stat file, write to it */
      if(statFileName!=NULL) {

	/* iteration number */
	fwrite(&iterations, sizeof(int), 1, statFile);

	/* training set stats */
	fwrite(&tAvgSumSqrErr, sizeof(double), 1, statFile);
	fwrite(&tMinSumSqrErr, sizeof(double), 1, statFile);
	fwrite(&tMaxSumSqrErr, sizeof(double), 1, statFile);
	fwrite(&tPercentCorrect, sizeof(double), 1, statFile);
	fwrite(tErrorBuckets, sizeof(double), T_MAX_LENGTH*T_MAX_LENGTH,
	       statFile);
	fwrite(tErrorDistrib, sizeof(int), 100, statFile);

	/* control set stats */
	fwrite(&cAvgSumSqrErr, sizeof(double), 1, statFile);
	fwrite(&cMinSumSqrErr, sizeof(double), 1, statFile);
	fwrite(&cMaxSumSqrErr, sizeof(double), 1, statFile);
	fwrite(&cPercentCorrect, sizeof(double), 1, statFile);
	fwrite(cErrorBuckets, sizeof(double), T_MAX_LENGTH*T_MAX_LENGTH,
	       statFile);
	fwrite(cErrorDistrib, sizeof(int), 100, statFile);
		
	/* now write out the weights and biases */
	fwriteNetState(statFile);

	/* flush to the file */
	fflush(statFile);
	
      }
    }
    
    iterations++;

    /* figure out stop criteria */
    if(stopType==ITER) {
      if(iterations==maxIter) keepGoing=0;
    } else {
      if(stopType==AVG) {
	if(tAvgSumSqrErr<maxErr) keepGoing=0;
      } else
	if(tMaxSumSqrErr<maxErr) keepGoing=0;
    
      /* if last numSame errors the same, exit */
      for(j=0,different=0;j<numSame;j++)
	if(lastErrs[0]!=lastErrs[j]) different++;
      if(!different) keepGoing=0;
      
    }  
    
  } while(keepGoing);

  /* print run statistics */
  printf("\nRun took %g seconds\n\n", ReadSysClock() - startTime);
  printf("Finished in %d Iterations.  Final Stats:\n"
	     "  Training Set:\n"
	     "\tSum Sqr Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n"
	     "  Control Set:\n"
	     "\tSum Sqr Error (Avg, Max, Min):\t% 4.3g % 4.3g % 4.3g\n\n",
	     iterations, tAvgSumSqrErr, tMaxSumSqrErr, tMinSumSqrErr,
	     cAvgSumSqrErr, cMaxSumSqrErr, cMinSumSqrErr);

  /* if there is a stat file and this is not a normal check point,
     write to it */
  if(statFileName!=NULL && iterations%checkFreq) {
    
    /* iteration number */
    fwrite(&iterations, sizeof(int), 1, statFile);
    
    /* training set stats */
    fwrite(&tAvgSumSqrErr, sizeof(double), 1, statFile);
    fwrite(&tMinSumSqrErr, sizeof(double), 1, statFile);
    fwrite(&tMaxSumSqrErr, sizeof(double), 1, statFile);
    fwrite(&tPercentCorrect, sizeof(double), 1, statFile);
    fwrite(tErrorBuckets, sizeof(double), T_MAX_LENGTH*T_MAX_LENGTH, statFile);
    fwrite(tErrorDistrib, sizeof(int), 100, statFile);
    
    /* control set stats */
    fwrite(&cAvgSumSqrErr, sizeof(double), 1, statFile);
    fwrite(&cMinSumSqrErr, sizeof(double), 1, statFile);
    fwrite(&cMaxSumSqrErr, sizeof(double), 1, statFile);
    fwrite(&cPercentCorrect, sizeof(double), 1, statFile);
    fwrite(cErrorBuckets, sizeof(double), T_MAX_LENGTH*T_MAX_LENGTH, statFile);
    fwrite(cErrorDistrib, sizeof(int), 100, statFile);
    
    /* now write out the weights and biases */
    fwriteNetState(statFile);
    
    /* flush to the file */
    fflush(statFile);
    
  }

  /* free buffers and close files */
  for(i=0;controlSetSequences[i]!=NULL;i++)
    free(controlSetSequences[i]);
  free(controlSetSequences);
  free(controlSet);

  for(i=0;trainingSetSequences[i]!=NULL;i++)
    free(trainingSetSequences[i]);
  free(trainingSetSequences);
  free(trainingSet);

  free(tErrorBuckets);
  free(tLengthBuckets);
  free(cErrorBuckets);
  free(cLengthBuckets);

  free(aOneWeights);
  free(aTwoWeights);
  free(aThreeWeights);
  free(aOutputWeights);

  free(bOneWeights);
  free(bTwoWeights);
  free(bThreeWeights);
  free(bOutputWeights);

  free(lastErrs);
  
  fclose(statFile);
  
  return 0;
  
}

