#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include "neural.h"

#define square(xVal) (xVal)*(xVal)

Neuron *inputs[2];
Neuron *hidden[2];
Neuron *output;

/*
 * ReadSysClock - returns the time in seconds and fractions of seconds
 *     as a double (for use only for calculating deltas)
 */
double
ReadSysClock ()
{

    struct timeval tv;

    (void) gettimeofday(&tv, NULL);
    return (tv.tv_sec +tv.tv_usec / 1000000.0);

} /* end of ReadSysClock */

int main(int argc, char *argv[]) {

  int i;
  int iterations=0;
  int alwaysApplyErrDeltas=0;
  double sumSqrErr;
  double maxErr=0.1;
  double startTime;
  double lrate=0.5;
  double momentum=0;

  /* some small random weights and biases to start with */
  double hiddenBiases[2]= { -1.2, -0.9 };
  double outputBias=-2.3;
  double hiddenWeights[4]= { 0.01, 0.02, 0.03, 0.04 };
  double outputWeights[2]= { 0.1, 0.09 };
  
  /* init with command line params */
  for(i=1;i<argc;i++) {
    
    if(!strcmp("-maxErr",argv[i])) {
      maxErr=atof(argv[++i]);
    } else if(!strcmp("-A",argv[i])) {
      alwaysApplyErrDeltas=1;
    } else if(!strcmp("-lrate",argv[i])) {
      lrate=atof(argv[++i]);
    } else if(!strcmp("-momentum",argv[i])) {
      momentum=atof(argv[++i]);
    } else {
      printf("\nSyntax: %s [-A] [-maxErr <val>] [-lrate <val>] "
	     "[-momentum val]\n\n"
	     "\t-maxErr\t\tspecifies the target error value\n"
	     "\t-lrate\t\tspecifies the learning rate\n"
	     "\t-momentum\tspecifies the momentum for learning\n"
	     "\t-A\t\tchange weights after each trial\n\n", argv[0]);
      exit(1);
    }

  }

  printf("\nRunning XOR Problem with:\n\tMaxErr=%g\tlrate=%g\t\tmomentum=%g"
	 "\n\tError Deltas applied once per %s\n\n", maxErr, lrate, momentum,
	 (alwaysApplyErrDeltas?"trial.":"trial set."));
  setLearnParams(lrate, momentum, lrate, momentum);
  
  /* create the neurons */
  for(i=0;i<2;i++) {
    inputs[i]=createNeuron(0);
    hidden[i]=createNeuron(&(hiddenBiases[i]));
  }
  
  output=createNeuron(&outputBias);
  
  /* create the connections */

  /* inputs to hidden node 0 */
  connectNeurons(inputs[0], hidden[0], &(hiddenWeights[0]));
  connectNeurons(inputs[1], hidden[0], &(hiddenWeights[1]));

  /* inputs to hidden node 1 */
  connectNeurons(inputs[0], hidden[1], &(hiddenWeights[2]));
  connectNeurons(inputs[1], hidden[1], &(hiddenWeights[3]));

  /* hidden nodes to output */
  connectNeurons(hidden[0], output, &(outputWeights[0]));
  connectNeurons(hidden[1], output, &(outputWeights[1]));

  /* Now train the network until error falls below maxErr */
  startTime=ReadSysClock();
  do {

    /* reset sumErr */
    sumSqrErr=0;
    
    /* first do '00 -> 0' */
    resetNet(output);
    
    setActivation(inputs[0], 0);
    setActivation(inputs[1], 0);

    sumSqrErr+= square(backProp(output, 0)); /* back prop & get error */
    calcDeltas(output);      /* calculate change in weights for this input */

    if(alwaysApplyErrDeltas)
      applyDeltas(output);     /* change the weights */

    /* now do '01 -> 1' */
    resetNet(output);
    
    setActivation(inputs[0], 0);
    setActivation(inputs[1], 1);

    sumSqrErr+= square(backProp(output, 1)); /* back prop & get error */
    calcDeltas(output);      /* calculate change in weights for this input */

    if(alwaysApplyErrDeltas)
      applyDeltas(output);     /* change the weights */

    /* now do '10 -> 1' */
    resetNet(output);
    
    setActivation(inputs[0], 1);
    setActivation(inputs[1], 0);

    sumSqrErr+= square(backProp(output, 1)); /* back prop & get error */
    calcDeltas(output);      /* calculate change in weights for this input */

    if(alwaysApplyErrDeltas)
      applyDeltas(output);     /* change the weights */
    
    /* now do '11 -> 0' */
    resetNet(output);
    
    setActivation(inputs[0], 1);
    setActivation(inputs[1], 1);

    sumSqrErr+= square(backProp(output, 0)); /* back prop & get error */
    calcDeltas(output);      /* calculate change in weights for this input */

    applyDeltas(output);     /* change the weights */
    
    iterations++;
    
  } while(sumSqrErr>maxErr);

  /* print run statistics */
  printf("Run took %g seconds\n\n", ReadSysClock() - startTime);
  printf("Finished in %d iterations with a sum squared error of %g\n\n",
	 iterations, sumSqrErr);
  
  /* now print the network outputs for the various patterns */
  
  /* first print '00 -> 0' */
  resetNet(output);
  
  setActivation(inputs[0], 0);
  setActivation(inputs[1], 0);
  
  printf("Input pattern 0,0 yields:\t%g\n", findOutput(output));
  
  /* now print '01 -> 1' */
  resetNet(output);
  
  setActivation(inputs[0], 0);
  setActivation(inputs[1], 1);
  
  printf("Input pattern 0,1 yields:\t%g\n", findOutput(output));
    
  /* now print '10 -> 1' */
  resetNet(output);
  
  setActivation(inputs[0], 1);
  setActivation(inputs[1], 0);
  
  printf("Input pattern 1,0 yields:\t%g\n", findOutput(output));
    
  /* now print '11 -> 0' */
  resetNet(output);
  
  setActivation(inputs[0], 1);
  setActivation(inputs[1], 1);
  
  printf("Input pattern 1,1 yields:\t%g\n\n", findOutput(output));

  /* print network biases and weights */
  printf("Weights and Biases:\n\n");
  printf("% 9.6g\t% 9.6g\t% 9.6g\t% 9.6g\n", hiddenWeights[0],
	 hiddenWeights[1], hiddenWeights[2], hiddenWeights[3]);
  printf("\t% 9.6g\t\t\t% 9.6g\n", hiddenBiases[0], hiddenBiases[1]);
  printf("\t\t% 9.6g\t% 9.6g\n", outputWeights[0], outputWeights[1]);
  printf("\t\t\t% 9.6g\n\n", outputBias);
    
  return 0;
  
}




